from __future__ import annotations

import sympy as sp
from sympy.abc import x, y

from ramanujantools import Matrix, Position
from ramanujantools.cmf import CMF


class FFbar(CMF):
    r"""
    Represents a 2D Conservative Matrix Field that was generated using the f, fbar construction:
    $a(x, y) = f(x, y) - \bar{f}(x+1, y) = f(x+1, y-1) - \bar{f}(x, y-1)$,
    $b(x) = f\bar{f}(x, 0) - f\bar{f}(0, 0) = f\bar{f}(x, y) - f\bar{f}(0, y)$,

    $Mx = \begin{pmatrix} 0, b(x) \cr 1, a(x, y) \end{pmatrix}$
    $My = \begin{pmatrix} \bar{f}(x, y), b(x) \cr 1, f(x, y) \end{pmatrix}$
    """

    @staticmethod
    def linear_condition(f, fbar) -> sp.Expr:
        r"""
        Returns the linear condition value for `f` and `fbar`.

        Functions $f(x, y), \bar{f}(x, y)$ satisfy the linear condition iff:
        $f(x+1, y-1) - \bar{f}(x, y-1) + \bar{f}(x+1, y) - f(x, y) = 0$
        """
        return sp.simplify(
            f.subs([[x, x + 1], [y, y - 1]])
            - fbar.subs(y, y - 1)
            + fbar.subs(x, x + 1)
            - f
        )

    @staticmethod
    def quadratic_condition(f, fbar) -> sp.Expr:
        r"""
        Returns the quadratic condition value for `f` and `fbar`.

        Functions $f(x, y), \bar{f}(x, y)$ satisfy the quadratic condition iff:
        $f\bar{f}(x, y) - f\bar{f}(x, 0) -f\bar{f}(0, y) + f\bar{f}(0, 0) = 0$,
        where $f\bar{f}(x, y) = f(x, y) \cdot \bar{f}(x, y)$
        """
        ffbar = sp.simplify(f * fbar)
        return sp.simplify(
            ffbar - ffbar.subs(x, 0) - ffbar.subs(y, 0) + ffbar.subs([[x, 0], [y, 0]])
        )

    def __init__(self, f, fbar):
        r"""
        Constructs an FFbar CMF.

        Args:
            f: the $f$ function
            fbar: the $\bar{f}$ function

        Raises:
            ValueError: if f and fbar do not satisfy the linear condition or the quadratic condition.
        """
        if FFbar.linear_condition(f, fbar) != 0:
            raise ValueError(
                f"given f and fbar do not satisfy the linear condition! f={f}, fbar={fbar}"
            )
        if FFbar.quadratic_condition(f, fbar) != 0:
            raise ValueError(
                f"given f and fbar do not satisfy the quadratic condition! f={f}, fbar={fbar}"
            )

        self.f = f
        """The f function of the FFbar CMF"""
        self.fbar = fbar
        """The fbar function of the FFbar CMF"""

        super().__init__(
            matrices={
                x: Matrix([[0, self.b()], [1, self.a()]]),
                y: Matrix([[self.fbar, self.b()], [1, self.f]]),
            },
            validate=False,
        )

    def __repr__(self):
        return f"FFbar({self.f}, {self.fbar})"

    def subs(self, substitutions: Position) -> FFbar:
        self._validate_axes_substitutions(substitutions)
        return FFbar(self.f.subs(substitutions), self.fbar.subs(substitutions))

    @staticmethod
    def A(f, fbar) -> sp.Expr:
        r"""
        Returns the $a(x, y)$ function as constructed in the ffbar construction:
        $a(x, y) = f(x, y) - \bar{f}(x+1, y) = f(x+1, y-1) - \bar{f}(x, y-1)$
        """
        return f - fbar.subs(x, x + 1)

    def a(self) -> sp.Expr:
        r"""
        Returns the $a(x, y)$ function for this FFbar CMF.
        """
        return FFbar.A(self.f, self.fbar)

    @staticmethod
    def B(f, fbar) -> sp.Expr:
        r"""
        Returns the $b(x)$ function as constructed in the ffbar construction:
        $b(x) = f\bar{f}(x, 0) - f\bar{f}(0, 0) = f\bar{f}(x, y) - f\bar{f}(0, y)$,
        where $f\bar{f}(x, y) = f(x, y) \cdot \bar{f}(x, y)$
        """
        ffbar_x_0 = (f * fbar).subs(y, 0)
        return sp.simplify(ffbar_x_0 - ffbar_x_0.subs(x, 0))

    def b(self) -> sp.Expr:
        r"""
        Returns the $b(x)$ function for this FFbar CMF.
        """
        return FFbar.B(self.f, self.fbar)



from __future__ import annotations

import sympy as sp
from sympy.abc import z

from ramanujantools import Matrix, Position
from ramanujantools.cmf import DFinite
from ramanujantools.cmf.d_finite import theta


class pFq(DFinite):
    r"""
    Represents the pFq CMF, derived from the differentiation property of generalized hypergeometric functions:
    https://en.wikipedia.org/wiki/Generalized_hypergeometric_function
    """

    def __init__(
        self,
        p: int,
        q: int,
        z: sp.Expr = z,
    ):
        r"""
        Constructs a pFq CMF.
        Args:
            p: The number of numerator parameters in the hypergeometric function
            q: The number of denominator parameters in the hypergeometric function
            z: If given, will attempt to construct the CMF for a specific z value.
        """
        self.p = p
        self.q = q
        self.z = sp.S(z)
        super().__init__(p, q, z)

    def __repr__(self) -> str:
        return f"pFq({self.p, self.q, self.z})"

    @staticmethod
    def x_axes(p: int) -> list[sp.Symbol]:
        return sp.symbols(f"x:{p}")

    @staticmethod
    def y_axes(q: int) -> list[sp.Symbol]:
        return sp.symbols(f"y:{q}")

    @classmethod
    def axes_and_signs(cls, p, q, z) -> dict[sp.Expr, bool]:
        x_axes = {x_i: True for x_i in pFq.x_axes(p)}
        y_axes = {y_i: False for y_i in pFq.y_axes(q)}
        return {**x_axes, **y_axes}

    @classmethod
    def differential_equation(cls, p, q, z) -> sp.Poly:
        return sp.Poly(
            sp.expand(
                theta * sp.prod(theta + y_i - 1 for y_i in pFq.y_axes(q))
                - z * sp.prod(theta + x_i for x_i in pFq.x_axes(p))
            ),
            theta,
        )

    @classmethod
    def construct_matrix(
        cls, theta_matrix: Matrix, axis: sp.Symbol, *args, **kwargs
    ) -> Matrix:
        eye = Matrix.eye(theta_matrix.rows)
        denom = axis - 1 if axis.name.startswith("y") else axis
        return theta_matrix / denom + eye

    def ascend(
        self, trajectory: Position, start: Position
    ) -> tuple[pFq, Position, Position]:
        r"""
        Returns a tuple of (CMF, trajectory, start), such that:
        1. The CMF is ascended, i.e, the CMF of _{p+1}F_{q+1}
        2. The start position contains two new more symbols
        3. The trajectory is padded with two zeros
        such that for any two choices of parameters $x_p, y_p$ such that $x_p - y_p \in \mathbb{Z}$,
        the ascended trajectory matrix contains all constants of the original,
        and the ascended delta is the same as the original (in a type-2 limit context).
        """
        ascended_cmf = pFq(
            self.p + 1,
            self.q + 1,
            self.z,
        )
        xp = sp.Symbol(f"x{self.p}")
        yq = sp.Symbol(f"y{self.q}")
        ascended_start = start + Position({xp: xp, yq: yq})
        ascended_trajectory = trajectory + Position({xp: 0, yq: 0})
        return (ascended_cmf, ascended_trajectory.sorted(), ascended_start.sorted())

    def subs(self, substitutions: Position) -> pFq:
        self._validate_axes_substitutions(substitutions)
        return pFq(
            self.p,
            self.q,
            self.z.subs(substitutions),
        )

    @staticmethod
    def predict_rank(p: int, q: int, z: sp.Expr):
        """
        Returns the rank of the CMF (i.e, the rank of its matrices).
        """
        N = max(p, q + 1)
        if z == 1 and p == q + 1:
            N -= 1
        return N

    @staticmethod
    def theta_derivative(expr: sp.Expr):
        r"""
        Returns the \theta derivative of an expression,
        which is defined as z * d/dz.
        """
        return z * sp.Derivative(expr, z).simplify()

    @staticmethod
    def state_vector(
        a_values: list[sp.Rational], b_values: list[sp.Rational], z_eval: sp.Expr = z
    ):
        r"""
        Returns the state vector of a pFq CMF in a specific point.
        The state vector is of length N, and the ith element is $\theta^i pFq(\bar{a}, \bar{b}, z)$,
        with a_values, b_values and z substituted in.
        """
        p = len(a_values)
        q = len(b_values)
        a_symbols = sp.symbols(f"a:{p}")
        b_symbols = sp.symbols(f"b:{q}")
        values = [sp.hyper(a_symbols, b_symbols, z)]
        for _ in range(1, pFq.predict_rank(p, q, z_eval)):
            values.append(pFq.theta_derivative(values[-1]))
        a_subs = Position.from_list(a_values, "a")
        b_subs = Position.from_list(b_values, "b")
        return sp.hyperexpand(
            Matrix(values).transpose().subs(a_subs | b_subs | {z: z_eval})
        )

    @staticmethod
    def contiguous_relation(point, anchor, z) -> Matrix:
        a_point, b_point = point
        a_anchor, b_anchor = anchor
        p = len(a_point)
        q = len(b_point)
        start = Position.from_list(a_anchor, "x") | Position.from_list(b_anchor, "y")
        end = Position.from_list(a_point, "x") | Position.from_list(b_point, "y")
        return pFq(p, q, z).work(start, end)

    @staticmethod
    def evaluate(
        a_values: list[sp.Rational], b_values: list[sp.Rational], z_eval: sp.Rational
    ) -> sp.Expr:
        """
        Evaluates symbolically the pFq function at a specific point.
        Levarages the pFq CMF to calculate the contiguous relations.
        Works by selecting a small anchor point for which we calculate the state vector
        using sympy, and then calculating the `work` matrix which represents the contiguous
        relations transformation up to our desired point.
        """
        a_values = [sp.S(value) for value in a_values]
        b_values = [sp.S(value) for value in b_values]
        a_anchor = [
            sp.sign(value) * (value - (value.floor() - 1)) for value in a_values
        ]
        b_anchor = [
            sp.sign(value) * (value - (value.floor() - 2)) for value in b_values
        ]
        vector = pFq.state_vector(a_anchor, b_anchor, z_eval)
        m = pFq.contiguous_relation((a_values, b_values), (a_anchor, b_anchor), z_eval)
        return (vector * m)[0]



import math

import sympy
from sympy import Poly, Symbol


class GenericPolynomial:
    @staticmethod
    def of_degree(
        deg: int, var_name: str, s: Symbol, monic: bool = False
    ) -> tuple[Poly, list[Symbol]]:
        r"""
        Creates and returns a generic polynomial of a single variable with degree 'deg' with a list of
        its generic coefficients.

        Example:
            from sympy.abc import x
            f, _ = GenericPolynomial.of_degree(deg=3, var_name='f_', s=x, monic=True)
            =>
            f = f_0 * x**0 + f_1 * x**1 + f_2 * x**2 + x**3
        """
        poly_vars = list(sympy.symbols(f"{var_name}:{deg + 1}"))
        poly = Poly(sum(poly_vars[i] * s**i for i in range(deg)), s)
        if monic:
            poly += s**deg
            poly_vars = poly_vars[:-1]
        else:
            poly += poly_vars[deg] * s**deg

        return poly, poly_vars

    @staticmethod
    def _sum_to(n: int, num_var: int):
        if num_var < 1:
            raise Exception(f"number of variables {num_var} must be at least 1")
        if num_var == 1:
            yield [n]
            return
        for i in range(n + 1):
            for sub_sum in GenericPolynomial._sum_to(n=n - i, num_var=num_var - 1):
                yield [i] + sub_sum

    @staticmethod
    def of_combined_degree(deg: int, var_name: str, variables: list[Symbol]):
        r"""
        Similar to method 'of_degree', except this is a polynomial in several variables, and we bound the combined
        degree, which is the sum of the powers of the different variables. For example combined(x**2 y**3) = 5.
        The name of the coefficient of x**m * y**k will be 'varname(m,k)'

        Example:
            from sympy.abc import x, y
            f, _ = GenericPolynomial.of_combined_degree(deg=2, var_name='f_', variables=[x, y])
            =>
            f = f_(0,0) + f_(1,0) * x + f_(0,1) * y + f_(2,0) * x**2 + f_(1,1) * x * y + f_(0,2) * y**2
        """
        poly = 0
        coefficients = []
        for combined_degree in range(deg + 1):
            for powers in GenericPolynomial._sum_to(
                n=combined_degree, num_var=len(variables)
            ):
                powers_str = ",".join([str(power) for power in powers])
                c = sympy.Symbol(f"{var_name}({powers_str})")
                monom = math.prod(s**power for s, power in zip(variables, powers))
                poly += c * monom
                coefficients.append(c)

        return poly, coefficients

    @staticmethod
    def symmetric_polynomials(*expressions) -> list:
        r"""
        Returns a list of all the symmetric polynomials in the given expressions. The k'th element is the
        sum of product of k distinct elements from the 'expressions' list.
        For example, for expressions=[x,y,z] returns 1, x+y+z, xy+yz+zx, xyz.
        """
        t = sympy.Symbol("__symmetric_polynomials")
        poly = math.prod([t + v for v in expressions], start=sympy.Poly(1, t))
        return poly.all_coeffs()

    @staticmethod
    def as_symmetric(polynomial, symm_symbols: list[Symbol], symm_var_name: str):
        r"""
        Given a polynomial in (symbols, symm_symbols), tries to find a new polynomial as a function
        of the symmetric polynomial in the symm_symbols.

        Example:
            x, y, z = sympy.symbols('x y z')
            p = (x+y)*(x*y)*z + z**2*(x**2 + y**2) + (x+y) + 1
            as_symmetric(p, [x,y], 's_') = s1*s2*z + z**2*(s1**2 - 2*s2) + s1 + 1
        """
        n = len(symm_symbols)
        symm_polynomial = 0
        symm = GenericPolynomial.symmetric_polynomials(*symm_symbols)
        s = sympy.symbols(f"{symm_var_name}:{n + 1}")

        monomial, coefficient = Poly(polynomial, *symm_symbols).LT()
        while coefficient != 0:
            ex = list(monomial.exponents) + [0]
            prod = coefficient
            for i in range(0, len(symm_symbols)):
                prod *= s[i + 1] ** (ex[i] - ex[i + 1])

            symm_polynomial += prod
            prod_subs = prod.subs({s[i]: symm[i] for i in range(1, n + 1)})
            polynomial -= prod_subs
            monomial, coefficient = Poly(polynomial, *symm_symbols).LT()

        return symm_polynomial



import sympy as sp
from sympy.abc import x, y

from ramanujantools import Matrix, Position
from ramanujantools.cmf import CMF, FFbar, known_cmfs

c0, c1, c2, c3 = sp.symbols("c:4")


def test_linear_condition():
    assert FFbar.linear_condition(x + y, x - y) == 0
    assert FFbar.linear_condition(2**x - 3**y, 2**x + 3**y) != 0


def test_quadratic_condition():
    assert FFbar.quadratic_condition(x**2 + x * y + y**2, x - y) == 0
    assert FFbar.quadratic_condition(x**2 + x * y + y**2, x - y + 1) != 0


def test_subs():
    substitutions = Position({c0: c0**2, c1: 3 - c1, c2: 0})
    cmf = known_cmfs.cmf1()
    assert FFbar(cmf.f.subs(substitutions), cmf.fbar.subs(substitutions)) == cmf.subs(
        substitutions
    )


def test_ffbar_construction():
    assert known_cmfs.cmf1() == CMF(
        matrices={
            x: Matrix(
                [
                    [0, -c0 * c2 + (c0 + c1 * x) * (c2 + c3 * x)],
                    [1, c0 + c1 * (x + y) - c2 - c3 * (x - y + 1)],
                ]
            ),
            y: Matrix(
                [
                    [c2 + c3 * (x - y), -c0 * c2 + (c0 + c1 * x) * (c2 + c3 * x)],
                    [1, c0 + c1 * (x + y)],
                ]
            ),
        }
    )



import pytest
import random
from sympy.abc import x, y

from ramanujantools import Matrix
from ramanujantools.cmf import CMF
from ramanujantools.cmf.known_cmfs import e, pi, zeta3
from ramanujantools.solvers import CoboundarySolver


known_cmf_list = [e(), pi(), zeta3()]


def verify_solution(mx1: Matrix, mx2: Matrix, solution: Matrix, deg: int):
    r"""
    Checks if the vector space of polynomial matrices m(x) of degree at most 'deg' satisfying
    mx1 * m(x+1) = m(x) * mx2
    contain the given solution.
    Note that m(x)=0 is always a solution.
    """
    result = CoboundarySolver.find_coboundary(mx1, mx2, deg, x)
    assert result is not None

    mm, variables = result
    # Check that the given solution appears in the vector space of solutions 'mm'
    assignment = CoboundarySolver.solve_polynomial_matrix(
        mm - solution, symbol=x, variables=variables
    )
    assert len(assignment) > 0


@pytest.mark.parametrize(
    "matrix",
    [
        Matrix([[1, x], [x**2, x + 5]]),
        Matrix([[1 + x + x**2, x], [x**5, x + 5 + x**3]]),
    ],
)
def test_self_coboundary(matrix: Matrix):
    """
    Every matrix is coboundary equivalent to itself via the identity matrix.
    """
    verify_solution(
        mx1=matrix,
        mx2=matrix,
        solution=Matrix(
            [[1, 0], [0, 1]]
        ),  # TODO : create Matrix.ID2() method which returns a 2x2 identity matrix
        deg=1,
    )


@pytest.mark.parametrize("cmf", known_cmf_list)
def test_2_lines_cmf_coboundary(cmf: CMF):
    # coboundary random two lines in a given cmf
    line = random.randint(1, 10)
    verify_solution(
        mx1=cmf.M(x)({y: line}),
        mx2=cmf.M(x)({y: line + 1}),
        solution=cmf.M(y)({y: line}),
        deg=5,
    )


@pytest.mark.parametrize("cmf", known_cmf_list)
def test_cmf_coboundary(cmf: CMF):
    r"""
    Given a family of parametrized matrices, we can find if they are part of a cmf
    """
    verify_solution(mx1=cmf.M(x), mx2=cmf.M(x)({y: y + 1}), solution=cmf.M(y), deg=6)


def test_specific_coboundary():
    mx1 = Matrix([[0, -(x**8)], [1, x**4 + (1 + x) ** 4]])
    mx2 = Matrix([[0, -(x**8)], [1, x**4 + (1 + x) ** 4 + 2 * (x**2 + (1 + x) ** 2)]])
    solution = Matrix(
        [[x**4 * (1 - 2 * x), -(x**8) * (1 + 2 * x)], [2 * x - 1, x**4 * (1 + 2 * x)]]
    )

    verify_solution(mx1=mx1, mx2=mx2, solution=solution, deg=10)


# TODO: Add negative tests, for matrices which are not polynomially coboundary equivalent



from sympy.abc import n
import sympy as sp

from ramanujantools import LinearRecurrence, Matrix
from ramanujantools.pcf import PCF


def f(c, index=n):
    return sp.Function(c)(n).subs({n: index})


def test_repr():
    expected = "LinearRecurrence([n, 1, 3 - n**2])"
    r = eval(expected)
    assert expected == repr(r)


def test_relation():
    expected = [1, n, n**2, n**3 - 7, 13 * n - 12]
    assert expected == LinearRecurrence(expected).relation


def test_matrix():
    relation = [-1, n, n + 1]
    assert Matrix([[0, n + 1], [1, n]]) == LinearRecurrence(relation).recurrence_matrix


def test_limit():
    initial_values = Matrix([[2, 3, 5], [7, 11, 13]])
    r = LinearRecurrence([1, n + 1, n**2])
    depths = [2, 3, 19, 101]
    assert r.limit(depths, 0, initial_values) == r.recurrence_matrix.limit(
        {n: 1}, depths, {n: 0}, initial_values
    )


def test_evaluate_solution_fibonacci():
    r = LinearRecurrence([-1, 1, 1])
    start = 33
    end = start + 5
    initial_values = Matrix([[1, 1]])
    assert r.evaluate_solution(initial_values, start, end) == [2, 3, 5, 8, 13]


def test_evaluate_solution_basis_single():
    r = LinearRecurrence([1, n, n**2, 3])
    start = 0
    end = 100
    initial_values = Matrix([[0, 1, 0]])
    matrix = r.recurrence_matrix.walk({n: 1}, end, {n: start})
    evaluation = r.evaluate_solution(initial_values, start, end)
    assert initial_values.dot(matrix.col(-1)) == evaluation[-1]


def test_evaluate_solution_basis_list_with_given_index():
    r = LinearRecurrence([n, -n, 5, n**2 + n + 1])
    start = 3
    end = 100
    initial_values = Matrix([[0, 0, 1]])
    matrices = r.recurrence_matrix.walk(
        {n: 1}, list(range(1, end - start + 1)), {n: start}
    )
    evaluations = r.evaluate_solution(initial_values, start, end)
    assert len(matrices) == len(evaluations)
    for i in range(len(evaluations)):
        assert initial_values.dot(matrices[i].col(-1)) == evaluations[i]


def test_evaluate_solution_generic():
    r = LinearRecurrence([2, -n + 1, n**3 + n, 14])
    start = 17
    end = 123
    initial_values = Matrix([[5, -4, 2]])
    matrices = r.recurrence_matrix.walk(
        {n: 1}, list(range(1, end - start + 1)), {n: start}
    )
    evaluations = r.evaluate_solution(initial_values, start, end)
    assert len(matrices) == len(evaluations)
    for i in range(len(evaluations)):
        assert initial_values.dot(matrices[i].col(-1)) == evaluations[i]


def test_inflate():
    r = LinearRecurrence([f("a"), f("b"), f("c"), f("d")])
    assert LinearRecurrence(
        [
            f("a"),
            f("b") * f("e"),
            f("c") * f("e") * f("e", n - 1),
            f("d") * f("e") * f("e", n - 1) * f("e", n - 2),
        ]
    ) == r.inflate(f("e"))


def test_deflate():
    r = LinearRecurrence([f("a"), f("b"), f("c"), f("d")])
    assert r.inflate(1 / f("e")) == r.deflate(f("e"))


def test_fold():
    r = LinearRecurrence([f("a"), f("b"), f("c")])
    assert LinearRecurrence(
        [
            f("a"),
            f("a", n - 1) * f("d") + f("b"),
            f("b", n - 1) * f("d") + f("c"),
            f("c", n - 1) * f("d"),
        ]
    ) == r.fold(f("d"))


def test_unfold_poly():
    r = LinearRecurrence([-1, n, n])
    multiplier = n - 3
    folded = r.fold(multiplier)
    assert [r] == folded.unfold_poly(multiplier)


def test_unfold():
    r = LinearRecurrence([-1, n, (n - 3) * (n + 7) * (n - 11)])
    multiplier = (n - 1) * (n + 2)
    folded = r.fold(multiplier)
    assert (r, sp.Poly(multiplier, n)) == folded.unfold()[0]


def test_fold_solution_space():
    r = LinearRecurrence([n, n + 1, (n + 17) ** 2, 3])
    initial_values = Matrix([[1, 2, 3]])
    start = 5
    end = 10
    solution = r.evaluate_solution(initial_values, start, end)
    multiplier = n**2 + n + 1
    folded = r.fold(multiplier)
    assert folded == r + multiplier * r._shift(1)
    folded_initial_values = Matrix.hstack(initial_values, Matrix([solution[0]]))
    shifted_solution = folded.evaluate_solution(folded_initial_values, start + 1, end)
    assert solution[1:] == shifted_solution


def test_unfold_deflate():
    r = LinearRecurrence([-1, n, n])
    multiplier = n - 17
    inflation = n + 1
    folded = r.fold(multiplier).inflate(inflation)
    solutions = folded.unfold(-1)
    recurrence, actual_multiplier = solutions[-1]
    assert r == recurrence.deflate(inflation)
    assert actual_multiplier == sp.Poly(multiplier * inflation, n)


def test_compose_solution_space_constants():
    r = LinearRecurrence([-1, 1, 1])
    initial_values = Matrix([[1, 1]])
    start = 0
    end = 5
    assert [2, 3, 5, 8, 13] == r.evaluate_solution(initial_values, start, end)

    rr = r.compose(r)
    new_initial_values = Matrix([[1, 1, 2, 3]])
    assert [5, 8, 13, 21, 34] == rr.evaluate_solution(new_initial_values, start, end)


def test_compose_solution_space_polynomials():
    r1 = LinearRecurrence([n**2 + 3 * n, 5 * n - 7, 13 * n**3, 2])
    r2 = LinearRecurrence([17 * n, 18 * (n - 2), 19 * (n - 3), 20 * (n - 5)])
    initial_values = Matrix([[17, 18, 19]])
    start = 17
    end = 100
    solution = r2.evaluate_solution(initial_values, start, end)

    rr = r1.compose(r2)
    shift = r1.order()
    composed_initial_values = Matrix.hstack(
        initial_values,
        Matrix([solution[:shift]]),
    )
    expected = solution[shift:]
    actual = rr.evaluate_solution(composed_initial_values, start + shift, end)

    assert expected == actual


def test_fold_is_compose():
    r = LinearRecurrence([n**2 + 3 * n, 5 * n - 7, 13 * n**3, 2])
    multiplier = 12 * n**3 - 34 * n**2 + 56 * n - 78
    assert r.fold(multiplier) == LinearRecurrence([1, multiplier]).compose(r)


def test_gamma():
    # from https://arxiv.org/abs/1010.1420
    m = Matrix([[0, -(n**2), 0], [1, 2 * n + 2, 0], [0, -(n - 1) / (n + 1), 1]])
    r = LinearRecurrence(m)
    assert [] == r.unfold()
    lim = r.limit(1000, 2)
    assert Matrix([[3, 12, 59], [6, 21, 102]]) == lim.identify(lim.mp.euler)
    assert Matrix([[1, 4, 20], [2, 7, 34]]) == lim.identify(
        -lim.mp.e * lim.mp.ei(-1)  # Gompertz
    )
    inflated = r.inflate(2 - n)

    lim = inflated.limit(1000, 5)
    assert Matrix([[3625, -26792, 461718], [6270, -46380, 799620]]) == lim.identify(
        lim.mp.euler, maxcoeff=10**6
    )

    unfolded = inflated.unfold_poly(n)
    assert 1 == len(unfolded)
    pcf = PCF(unfolded[0].recurrence_matrix).deflate_all()
    assert PCF(-2 * (n + 2), -((n + 1) ** 2)) == pcf


def test_euler_gompertz_independent():
    # from https://arxiv.org/abs/0902.1768
    r = LinearRecurrence(
        [
            -(16 * n + 1) * (16 * n - 15),
            (16 * n - 15) * (256 * n**3 + 528 * n**2 + 352 * n + 73),
            -(16 * n + 17) * (128 * n**3 + 40 * n**2 - 82 * n - 45),
            n**2 * (16 * n + 17) * (16 * n + 1),
        ]
    )

    def gompertz(mp):
        return -mp.e * mp.ei(-1)

    def constant(mp):
        return mp.e * mp.euler + gompertz(mp)

    limit = r.limit(200, 1)
    assert Matrix([[0, 0, 17], [-3, -2, 7]]) == limit.identify(constant(limit.mp))



import sympy as sp
from sympy.abc import x, y, z

from ramanujantools import Position


def test_add():
    p1 = Position({x: 17, y: 3})
    p2 = Position({y: 4, z: 3})
    assert {x: 17, y: 7, z: 3} == p1 + p2


def test_iadd():
    p = Position({x: 17, y: 3})
    p += Position({y: 4, z: 3})
    assert {x: 17, y: 7, z: 3} == p


def test_isub():
    p = Position({x: 17, y: 3})
    p -= Position({y: 4, z: 3})
    assert {x: 17, y: -1, z: -3} == p


def test_sub():
    p1 = Position({x: 17, y: 3})
    p2 = Position({y: 4, z: 3})
    assert {x: 17, y: -1, z: -3} == p1 - p2


def test_mul():
    p = Position({x: 1, y: -2, z: 3})
    expected = {key: -7 * value for key, value in p.items()}
    assert expected == -7 * p
    assert expected == p * -7


def test_neg():
    p = Position({x: 1, y: -2, z: 3})
    assert {key: -value for key, value in p.items()} == -p


def test_longest():
    p = Position({x: 1, y: 2, z: -3})
    assert 3 == p.longest()


def test_longest_empty():
    p = Position()
    assert 0 == p.longest()


def test_shortest():
    p = Position({x: 1, y: 2, z: -3})
    assert 1 == p.shortest()


def test_shortest_empty():
    p = Position()
    assert 0 == p.longest()


def test_signs():
    p = Position({x: 1, y: 2, z: -3})
    assert Position({x: 1, y: 1, z: -1}) == p.signs()


def test_signs_empty():
    assert Position() == Position().signs()


def test_is_integer():
    assert Position({x: 1, y: 7}).is_integer()
    assert not Position({x: 1, y: x**2 + 3 * x + 1}).is_integer()
    assert not Position({x: x / 2, y: x**2 + 3 * x + 1}).is_integer()
    assert not Position({x: sp.Rational(1, 2), y: 1}).is_integer()


def test_is_polynomial():
    assert Position({x: 1, y: 7}).is_polynomial()
    assert Position({x: 1, y: x**2 + 3 * x + 1}).is_polynomial()
    assert not Position({x: x / 2, y: x**2 + 3 * x + 1}).is_polynomial()
    assert not Position({x: sp.Rational(1, 2), y: 1}).is_polynomial()


def test_is_denominator_lcm():
    assert 1 == Position({x: 1, y: 7}).denominator_lcm()
    assert 1 == Position({x: 1, y: x**2 + 3 * x + 1}).denominator_lcm()
    assert 2 == Position({x: x / 2, y: x**2 + 3 * x + 1}).denominator_lcm()
    assert 15 == Position({x: sp.Rational(1, 3), y: x / 5}).denominator_lcm()


def test_free_symbols():
    assert {x, z} == Position({x: 1, y: z, z: x}).free_symbols()


def test_from_list():
    z0, z1, z2, z3 = sp.symbols("z:4")
    expected = Position({z0: 9, z1: 8, z2: 7, z3: 6})
    actual = Position.from_list([9, 8, 7, 6], "z")
    assert isinstance(actual, Position)
    assert expected == actual



from __future__ import annotations

from abc import ABC, abstractmethod

import sympy as sp

from ramanujantools import Matrix
from ramanujantools.cmf import CMF

from functools import lru_cache


theta = sp.symbols("theta")


class DFinite(CMF, ABC):
    """
    Represents a D-finite CMF, which is a base class for constructing CMFs that are derived from D-finite functions.
    In order to implement a D-finite CMF, one must define the following methods:
    1. axes: Returns the axes and their signs (directions) for the D-finite CMF.
    2. differential_equation: Returns the differential equation for the D-finite CMF.
    3. construct_matrix: Given a symbol, constructs the matrix for the D-finite CMF.
    """

    def __init__(self, *args, **kwargs):
        """
        Constructs the D-finite CMF.
        The D-finite CMF is a base class for constructing CMFs that are derived from D-finite functions.
        """
        matrices, negative_matrices = self.construct_matrices(*args, **kwargs)
        super().__init__(
            matrices=matrices,
            _negative_matrices_cache=negative_matrices,
            validate=False,
        )

    @classmethod
    @abstractmethod
    def axes_and_signs(cls, *args, **kwargs) -> dict[sp.Expr, bool]:
        """
        Function returns the axes and their signs for the D-finite CMF.
        The signs are used to determine the direction of the axes in the CMF,
        and are derived from the contiguous relations of the D-finite function.
        """
        pass

    @classmethod
    @abstractmethod
    def differential_equation(cls, *args, **kwrags) -> sp.Poly:
        """
        Returns the differential equation for the D-finite CMF.
        """
        pass

    @classmethod
    @abstractmethod
    def construct_matrix(
        cls, theta_matrix: Matrix, axis: sp.Symbol, *args, **kwargs
    ) -> Matrix:
        """
        Given a symbol, constructs the matrix for the D-finite CMF.
        Note that if axes[symbol] is False, then the matrix should be used to decrement the axis by one.
        """
        pass

    @classmethod
    @lru_cache
    def construct_matrices(
        cls, *args, **kwargs
    ) -> tuple[dict[sp.Expr, Matrix], dict[sp.Expr, Matrix]]:
        """
        Constructs the matrices for the D-finite CMF.
        """
        d_poly = cls.differential_equation(*args, **kwargs)

        if not (d_poly.free_symbols).issuperset(
            cls.axes_and_signs(*args, **kwargs).keys()
        ):
            raise ValueError(
                "The differential equation free symbols does not contain the axes!"
            )

        d_poly_monic = sp.Poly(d_poly / sp.LC(d_poly), theta)
        theta_matrix = Matrix.companion(d_poly_monic)

        matrices = {}
        negative_matrices = {}
        for axis, sign in cls.axes_and_signs(*args, **kwargs).items():
            matrix = cls.construct_matrix(theta_matrix, axis, *args, **kwargs)
            if sign:
                matrices[axis] = matrix
            else:
                negative_matrices[axis] = matrix
                matrices[axis] = matrix.subs({axis: axis + 1}).inverse().factor()
        return matrices, negative_matrices



import sympy as sp
from sympy.abc import n, z

from ramanujantools.cmf import CMF
from ramanujantools.cmf import pFq
from ramanujantools import Matrix, Position

x0, x1, x2, x3 = sp.symbols("x:4")
y0, y1, y2 = sp.symbols("y:3")


def test_subs():
    cmf = pFq(2, 1)
    assert pFq(2, 1, -1) == cmf.subs({z: -1})


def test_ascend_symbolic():
    cmf = pFq(2, 1)
    start = Position({x0: 1, x1: 2, y0: 3})
    trajectory = Position({x0: 4, x1: 5, y0: 6})

    expected_cmf = pFq(3, 2)
    expected_start = start + Position({x2: x2, y1: y1})
    expected_trajectory = trajectory + Position({x2: 0, y1: 0})

    assert (expected_cmf, expected_trajectory, expected_start) == cmf.ascend(
        trajectory, start
    )


def test_ascend_zeta2():
    cmf = pFq(3, 2, 1)
    trajectory = Position({x0: 1, x1: 1, x2: 1, y0: 2, y1: 2})
    start = 2 * Position({x0: 1, x1: 1, x2: 1, y0: 2, y1: 2})

    limit = cmf.trajectory_matrix(trajectory, start).limit({n: 1}, 200, {n: 1})
    assert limit.identify(limit.mp.zeta(2)) is not None
    delta = limit.delta(limit.mp.zeta(2))

    a_cmf, a_trajectory, a_start = cmf.ascend(trajectory, start)
    a_limit = (
        a_cmf.trajectory_matrix(a_trajectory, a_start)
        .subs({x3: 3, y2: 3})
        .limit({n: 1}, 200, {n: 1})
    )
    assert a_limit.identify(a_limit.mp.zeta(2), maxcoeff=10**6) is not None
    a_delta = a_limit.delta(a_limit.mp.zeta(2))

    assert abs(delta - a_delta) < 0.01


def test_2F1():
    x0 = sp.Symbol("x0")
    x1 = sp.Symbol("x1")
    y0 = sp.Symbol("y0")
    z = sp.Symbol("z")
    expected = CMF(
        {
            x0: Matrix(
                [
                    [1, -x1 * z / (z - 1)],
                    [1 / x0, 1 - (x0 * z + x1 * z - y0 + 1) / (x0 * (z - 1))],
                ]
            ),
            x1: Matrix(
                [
                    [1, -x0 * z / (z - 1)],
                    [1 / x1, 1 - (x0 * z + x1 * z - y0 + 1) / (x1 * (z - 1))],
                ]
            ),
            y0: Matrix(
                [
                    [
                        y0 * (-x0 - x1 + y0) / (x0 * x1 - x0 * y0 - x1 * y0 + y0**2),
                        x0 * x1 * y0 / (x0 * x1 - x0 * y0 - x1 * y0 + y0**2),
                    ],
                    [
                        y0 * (1 - z) / (z * (x0 * x1 - x0 * y0 - x1 * y0 + y0**2)),
                        y0**2 * (z - 1) / (z * (x0 * x1 - x0 * y0 - x1 * y0 + y0**2)),
                    ],
                ]
            ),
        }
    )
    cmf = pFq(2, 1)
    cmf.validate_conserving()
    assert cmf == expected


def test_2F1_z_evaluation():
    p = 2
    q = 1
    z_value = -7
    assert pFq(p, q, z=z_value) == pFq(p, q).subs({z: z_value})


def test_gamma():
    cmf = pFq(2, 2, z=-1)
    x0, x1 = sp.symbols("x:2")
    y0, y1 = sp.symbols("y:2")
    trajectory = {x0: 1, x1: 1, y0: 0, y1: -1}
    start = {x0: 1, x1: 1, y0: -1, y1: -1}
    limit = cmf.limit(trajectory, 100, start)
    assert Matrix([[1, 3, 0], [3, 5, 0]]) == limit.identify(limit.mp.euler)


def test_pfq_conserving():
    for p in range(1, 3):
        for q in range(1, 3):
            pFq(p, q).validate_conserving()


def test_predict_rank():
    for p in range(1, 5):
        for q in range(1, 5):
            for _z in [-1, 1]:
                assert pFq(p, q, z=_z).rank() == pFq.predict_rank(p, q, _z)


def test_work_numeric_3f2():
    cmf = pFq(3, 2, -1)
    x0, x1, x2 = sp.symbols("x:3")
    y0, y1 = sp.symbols("y:2")
    start = Position(
        {x0: 1, x1: -sp.Rational(1, 2), x2: 0, y0: sp.Rational(5, 3), y1: -1}
    )
    end = Position(
        {x0: 17, x1: -sp.Rational(13, 2), x2: 0, y0: sp.Rational(20, 3), y1: -4}
    )
    trajectory = end - start
    expected = cmf.walk(trajectory, 1, start)
    actual = cmf.work(start, end)
    assert expected == actual


def test_state_vector():
    p = 3
    q = 2
    z_eval = -1
    a_symbols = sp.symbols(f"a:{p}")
    b_symbols = sp.symbols(f"b:{q}")
    state_vector = pFq.state_vector(a_symbols, b_symbols, z_eval)
    assert len(state_vector) == pFq.predict_rank(p, q, z_eval)
    assert state_vector.free_symbols == set(a_symbols).union(b_symbols)
    value = sp.hyper(a_symbols, b_symbols, z)
    for i in range(len(state_vector)):
        assert sp.hyperexpand(value.subs({z: z_eval})) == state_vector[i]
        value = pFq.theta_derivative(value)


def test_evaluate_2f1():
    for _z in [
        -1,
        -sp.Rational(1, 2),
        -sp.Rational(1, 4),
        sp.Rational(1, 4),
        sp.Rational(1, 2),
    ]:
        for a_values, b_values in [
            ([1, 1], [2]),
            ([1, 3], [5]),
            ([sp.Rational(3, 2), sp.Rational(5, 2)], [sp.Rational(7, 2)]),
            ([sp.Rational(3, 2), 4], [-sp.Rational(1, 2)]),
        ]:
            evaluation = pFq.evaluate(a_values, b_values, _z).simplify()
            expected = sp.hyper(a_values, b_values, _z).simplify().simplify()
            assert expected == evaluation



from __future__ import annotations

import sympy as sp
from sympy.abc import z

from ramanujantools import Matrix
from ramanujantools.cmf import DFinite
from ramanujantools.cmf.d_finite import theta


class MeijerG(DFinite):
    r"""
    Represents the pFq CMF, derived from the differentiation property of generalized hypergeometric functions:
    https://en.wikipedia.org/wiki/Meijer_G-function
    """

    def __init__(self, m: int, n: int, p: int, q: int, z: sp.Expr = z):
        if not (p > 0 and 0 <= n and n <= p) or not (q > 0 and 0 <= m and m <= q):
            raise ValueError(
                "Meijer G must satisfy p > 0, 0 <= n <= p and q > 0, 0 <= m <= q"
            )
        self.m = m
        self.n = n
        self.p = p
        self.q = q
        self.z = z
        super().__init__(self.m, self.n, self.p, self.q, self.z)

    def __repr__(self) -> str:
        return f"MeijerG({self.m, self.n, self.p, self.q, self.z})"

    @staticmethod
    def a_axes(p) -> list[sp.Symbol]:
        return sp.symbols(f"a:{p}")

    @staticmethod
    def b_axes(q) -> list[sp.Symbol]:
        return sp.symbols(f"b:{q}")

    @classmethod
    def axes_and_signs(cls, m, n, p, q, z) -> dict[sp.Expr, bool]:
        a_axes = {a_i: False for a_i in MeijerG.a_axes(p)}
        b_axes = {b_i: True for b_i in MeijerG.b_axes(q)}
        return {**a_axes, **b_axes}

    @staticmethod
    def differential_equation(m, n, p, q, z) -> sp.Poly:
        coeff = (-sp.Integer(1)) ** (p - m - n)
        return sp.Poly(
            sp.expand(
                coeff * z * sp.prod(theta - a_i + 1 for a_i in MeijerG.a_axes(p))
                - sp.prod(theta - b_j for b_j in MeijerG.b_axes(q))
            ),
            theta,
        )

    @classmethod
    def construct_matrix(
        cls, theta_matrix: Matrix, axis: sp.Symbol, m: int, n: int, *args, **kwargs
    ) -> Matrix:
        eye = Matrix.eye(theta_matrix.rows)
        is_a = axis.name.startswith("a")
        index = int(axis.name[1:])
        if is_a:
            sign = -1 if index > n else 1
        else:
            sign = 1 if index > m else -1
        multiplier = axis - 1 if is_a else axis
        return (theta_matrix - multiplier * eye) * sign



from pytest import approx

from unittest.mock import patch

import sympy as sp
from sympy.abc import x, y, n

from ramanujantools import Matrix, Limit, simplify
from ramanujantools.pcf import PCF
from ramanujantools.cmf import pFq


def test_eq_optimized_path_called():
    m1 = Matrix.eye(2)
    m2 = Matrix.eye(2)
    with patch("ramanujantools.Matrix._optimized_eq", return_value=True) as mock_opt:
        assert m1 == m2
        mock_opt.assert_called_once_with(m1, m2)


def test_eq_non_polynomial():
    assert not Matrix.eye(2) == (2**x) * Matrix.eye(2)


def test_is_square():
    assert Matrix([[1, 2], [3, 4]]).is_square()
    assert not Matrix([1, 2, 3, 4]).is_square()


def test_gcd():
    a = 2 * 3 * 5
    b = 2 * 3 * 7
    c = 2 * 5 * 7
    d = 3 * 5 * 7
    m = Matrix([[a, b], [c, d]])
    m *= 11
    assert 11 == m.gcd


def test_reduce():
    initial = Matrix([[2, 3], [5, 7]])
    gcd = sp.Rational(17, 13)
    m = gcd * initial
    assert m.gcd == gcd
    assert m.reduce() == initial


def test_is_numeric_walk():
    m = Matrix([[x, 1], [y, 2]])

    assert m._is_numeric_walk({x: 1, y: 1}, {x: 1, y: 1})
    assert m._is_numeric_walk(
        {x: 1, y: 1}, {x: sp.Rational(1, 2), y: sp.Rational(1, 2)}
    )

    # x remains a symbol
    assert not m._is_numeric_walk({x: 1, y: 1}, {x: x, y: 1})

    # rational coefficients of symbolic polynomials are supported
    assert not m._is_numeric_walk({x: 1, y: 1}, {x: x / 2, y: 1})

    # substituting a different symbol causes symbolic flint calculation
    assert not m._is_numeric_walk({x: 1, y: 1}, {x: n - 1, y: n**2})


def test_subs_degenerated():
    m = Matrix([[x, 1], [y, 2]])
    assert m == m.subs({x: x})
    assert m == m.subs({y: y})
    assert m == m.subs({x: x, y: y})


def test_subs_numerical():
    m = Matrix([[x, x**2], [13 + x, -x]])
    substitutions = {x: 5}
    assert Matrix([[5, 25], [18, -5]]) == m.subs(substitutions)


def test_subs_symbolic():
    m = Matrix([[x, x**2, 13 + x, -x]])
    expr = y**2 + x - 3
    assert Matrix([[expr, (expr) ** 2, 13 + expr, -expr]]) == m.subs({x: expr})


def test_as_polynomial():
    m = Matrix([[1, 1 / x], [0, 3 / (x**2 - x)]])
    polynomial_m = Matrix([[x * (x - 1), x - 1], [0, 3]])
    assert polynomial_m == m.as_polynomial()


def test_factor_symbolic():
    matrix = Matrix(
        [
            [x**3 + 3 * x**2 + 3 * x + 1, x**100 - x**90 + x**80],
            [(x + 1) ** 2, 3 / (x**2 - x)],
        ]
    )
    assert matrix.applyfunc(sp.factor) == matrix.factor()


def test_factor_numeric():
    matrix = Matrix(
        [[0, -1, 2, -3], [-4, 5, -6, 7], [-8, 9, 10, -11], [12, -13, -14, 15]]
    )
    assert matrix.applyfunc(sp.factor) == matrix.factor()


def test_inverse():
    a = 5
    b = 2
    c = 3
    d = 7
    m = Matrix([[a, b], [c, d]])
    expected = Matrix([[d, -b], [-c, a]]) / (a * d - b * c)
    assert expected == m.inverse()


def test_singular_points_nonvariable():
    m = Matrix([[1, 2], [3, 4]])
    assert len(m.singular_points()) == 0


def test_singular_points_single_variable():
    m = Matrix([[1, 0], [1, (x - 1) * (x - 3)]])
    assert m.singular_points() == [{x: 1}, {x: 3}]


def test_singular_points_multi_variable():
    m = Matrix([[1, x], [1, y]])
    assert m.singular_points() == [{x: y}]


def test_coboundary():
    m = Matrix([[1, n, 2], [3, n**2, 5 * n], [n - 7, n**2 + 1, n - 3]])
    U = Matrix([[3, 1, n - 2], [5, n**2 - 3 * n + 1, 0], [11 * n + 2, 3, n - 19]])
    expected = U.inverse() * m * U.subs({n: n + 1})
    assert expected == m.coboundary(U)


def test_coboundary_inverse():
    m = Matrix([[1, n, 2], [3, n**2, 5 * n], [n - 7, n**2 + 1, n - 3]])
    U = Matrix([[3, 1, n - 2], [5, n**2 - 3 * n + 1, 0], [11 * n + 2, 3, n - 19]])
    assert m == m.coboundary(U).coboundary(U.inverse())


def test_companion_form():
    expected = Matrix([[0, 0, n], [1, 0, 17], [0, 1, n**2]])
    assert expected == Matrix.companion_form([n, 17, n**2])


def test_is_companion():
    assert not Matrix.eye(3).is_companion()
    m = Matrix([[0, 0, n**3 - 1], [1, 0, n**2 + 3], [0, 1, 2 * n]])
    assert m.is_companion()
    assert not (m + Matrix.eye(3)).is_companion()


def test_companion_coboundary():
    m = Matrix([[1, n, 2], [3, n**2, 5 * n], [n - 7, n**2 + 1, n - 3]])
    assert m.coboundary(m.companion_coboundary_matrix()).is_companion()


def test_as_companion():
    m = Matrix(
        [
            [1, -1, -n * (1 - 1 / n) - n - 1],
            [2 / n, 1 - 2 / n, -2 + (-2 * n - 1) / n + (-n - 1) / n + 2 / n],
            [
                n ** (-2),
                (1 - 1 / n) / n + 1 / n,
                (1 - 1 / n) ** 2 + (-2 * n - 1) / n**2,
            ],
        ]
    )

    companion = m.as_companion()
    assert companion.is_companion()


def test_as_companion_smaller_recursion():
    # This matrix is derived from pFq(4, 3, 1) with
    # start = {x0: 0, x1: 0, x2: 1, x3: 1, y0: 1, y1: 1, y2: 1},
    # trajectory = {x0: -1, x1: -1, x2: 1, x3: 1, y0: 0, y1: 0, y2: 0}
    m = Matrix(
        [
            [(17 * n - 12) / n, 12 * n - 8, 4 * n * (2 * n - 1)],
            [(24 * n - 24) / n**2, (17 * n - 16) / n, 12 * n - 8],
            [-12 / n**3, -8 / n**2, (n - 4) / n],
        ]
    )
    companion = m.as_companion()
    assert companion.is_companion()
    assert 2 == companion.rows
    # This is Apery's PCF
    assert PCF(34 * n**3 + 51 * n**2 + 27 * n + 5, -(n**6)) == PCF(companion)


def test_companion_coboundary_two_variables():
    m = Matrix([[1, x, 2 * y], [3, x * y, 5 * y], [x - 7, (x + y) ** 2 + 1, y - 3]])
    assert m.coboundary(m.companion_coboundary_matrix(x), x).is_companion()
    assert m.coboundary(m.companion_coboundary_matrix(y), y).is_companion()


def test_walk_0():
    m = Matrix([[x, 3 * x + 5 * y], [y**7 + x - 3, x**5]])
    assert m.walk({x: 0, y: 1}, 0, {x: x, y: y}) == Matrix.eye(2)


def test_walk_1():
    m = Matrix([[x, 3 * x + 5 * y], [y**7 + x - 3, x**5]])
    assert m.walk({x: 1, y: 0}, 1, {x: x, y: y}) == m


def test_walk_list():
    trajectory = {x: 2, y: 3}
    start = {x: 5, y: 7}
    iterations = [1, 2, 3, 17, 29, 53, 99]
    m = Matrix([[x, 3 * x + 5 * y], [y**7 + x - 3, x**5]])
    assert m.walk(trajectory, iterations, start) == [
        m.walk(trajectory, i, start) for i in iterations
    ]


def test_walk_start_single_variable():
    iterations = [1, 2, 3, 4]
    m = Matrix([[0, x**2], [1, x + 1]])
    expected = m.walk({x: 1}, sum(iterations), {x: 1})
    actual = Matrix.eye(2)
    for i in range(len(iterations)):
        actual *= m.walk({x: 1}, iterations[i], {x: 1 + sum(iterations[0:i])})
    assert expected == actual


def test_walk_start_multi_variable():
    iterations = [1, 2, 3, 4]
    m = Matrix([[0, x**2], [1, y + 1]])
    starting_point = {x: 2, y: 3}
    trajectory = {x: 5, y: 7}
    expected = m.walk(trajectory, sum(iterations), starting_point)
    actual = Matrix.eye(2)
    for i in range(len(iterations)):
        actual *= m.walk(
            trajectory,
            iterations[i],
            {
                x: starting_point[x] + sum(iterations[0:i]) * trajectory[x],
                y: starting_point[y] + sum(iterations[0:i]) * trajectory[y],
            },
        )
    assert expected == actual


def test_walk_axis():
    m = Matrix([[x, 3 * x + 5 * y], [y**7 + x - 3, x**5]])
    assert simplify(m.walk({x: 1, y: 0}, 3, {x: 1, y: 1})) == simplify(
        m({x: 1, y: 1}) * m({x: 2, y: 1}) * m({x: 3, y: 1})
    )
    assert simplify(m.walk({x: 0, y: 1}, 5, {x: 1, y: 1})) == simplify(
        m({x: 1, y: 1})
        * m({x: 1, y: 2})
        * m({x: 1, y: 3})
        * m({x: 1, y: 4})
        * m({x: 1, y: 5})
    )


def test_walk_diagonal():
    m = Matrix([[x, 3 * x + 5 * y], [y**7 + x - 3, x**5]])
    assert simplify(m.walk({x: 1, y: 1}, 4, {x: 1, y: 1})) == simplify(
        m({x: 1, y: 1}) * m({x: 2, y: 2}) * m({x: 3, y: 3}) * m({x: 4, y: 4})
    )
    assert simplify(m.walk({x: 3, y: 2}, 3, {x: 1, y: 1})) == simplify(
        m({x: 1, y: 1}) * m({x: 4, y: 3}) * m({x: 7, y: 5})
    )


def test_walk_different_start():
    m = Matrix([[x, 3 * x + 5 * y], [y**7 + x - 3, x**5]])
    assert simplify(m.walk({x: 3, y: 2}, 3, {x: 5, y: 7})) == simplify(
        m({x: 5, y: 7}) * m({x: 8, y: 9}) * m({x: 11, y: 11})
    )


def test_walk_equivalent_to_limit():
    trajectory = {x: 2, y: 3}
    start = {x: 5, y: 7}
    iterations = 17
    m = Matrix([[x, 3 * x + 5 * y], [y**7 + x - 3, x**5]])
    m_previous, m_last = m.walk(trajectory, [iterations - 1, iterations], start)
    assert m.limit(trajectory, iterations, start) == Limit(m_last, m_previous)


def test_limit_list():
    trajectory = {x: 2, y: 3}
    start = {x: 5, y: 7}
    iterations = [1, 2, 3, 17, 29, 53, 99]
    m = Matrix([[x, 3 * x + 5 * y], [y**7 + x - 3, x**5]])
    limits = m.limit(trajectory, iterations, start)
    for depth_index in range(len(iterations)):
        assert limits[depth_index] == m.limit(
            trajectory, iterations[depth_index], start
        )


def test_limit_initial_values():
    trajectory = {x: 2, y: 3}
    start = {x: 5, y: 7}
    iterations = 10
    m = Matrix([[x, 3 * x + 5 * y], [y**7 + x - 3, x**5]])
    initial_values = Matrix([[2, 3, 5], [7, 11, 13]])
    limit = m.limit(trajectory, iterations, start, initial_values)
    assert initial_values == limit.initial_values
    assert limit != m.limit(trajectory, iterations, start)


def test_charpoly():
    m = Matrix([[0, -(n**2)], [1, (3 * n + 1)]])
    assert sp.Matrix(m).charpoly() == m.charpoly(poincare=False)


def test_poincare_poly():
    poly = sp.PurePoly(2 * x**3 + x**2 + x * (4 * n**2 + 2) + (n**3), x)
    expected = sp.PurePoly(2 * x**3 + 4 * x + 1, x)
    assert expected == Matrix.poincare_poly(poly)


def test_poincare_poly_degenerated():
    poly = sp.PurePoly(2 * x**3 + x**2 + x * (4 * n**2 + 2) + (n**4), x)
    expected = sp.PurePoly(2 * x**3, x)
    assert expected == Matrix.poincare_poly(poly)


def test_poincare_poly_aptekarev():
    poly = sp.PurePoly(
        x**3
        + (-256 * n**3 - 528 * n**2 - 352 * n - 73) / (16 * n + 1) * x**2
        + (2048 * n**4 + 2816 * n**3 - 632 * n**2 - 2114 * n - 765)
        / (256 * n**2 - 224 * n - 15)
        * x
        + (-16 * n**3 - 17 * n**2) / (16 * n - 15),
        x,
    )
    expected = sp.PurePoly(x**3 - 16 * x**2, x)
    assert expected == Matrix.poincare_poly(poly)


def test_poincare_poly_constant():
    poly = sp.PurePoly(5 * x**3 + 2 * x**2 + x - 7, x)
    assert poly == Matrix.poincare_poly(poly)


def test_errors():
    x0, x1 = sp.symbols("x:2")
    (y0,) = sp.symbols("y:1")
    trajectory = {x0: 1, x1: 1, y0: 1}
    start = trajectory
    m = pFq(2, 1, -1).trajectory_matrix(trajectory, start)
    lambdas = m.sorted_eigenvals()
    assert sp.log(abs(lambdas[0]).evalf() / abs(lambdas[1]).evalf()) == approx(
        m.errors()[0]
    )


def test_gcd_slope():
    x0, x1 = sp.symbols("x:2")
    (y0,) = sp.symbols("y:1")
    trajectory = {x0: 1, x1: 1, y0: 1}
    start = {x0: 0, x1: 0, y0: 0}
    m = pFq(2, 1, -1).trajectory_matrix(trajectory, start)
    assert m.gcd_slope(20) == approx(1.3962425331281643)
    assert m.gcd_slope(40) == approx(1.5535146470266243)
    assert m.gcd_slope(100) == approx(1.5895449981095848)


def test_kamidelta_2f1():
    x0, x1 = sp.symbols("x:2")
    (y0,) = sp.symbols("y:1")
    trajectory = {x0: 1, x1: 2, y0: 3}
    start = trajectory
    m = pFq(2, 1, -1).trajectory_matrix(trajectory, start)
    actual = m.kamidelta()[0]
    l1, l2 = m.limit({n: 1}, [100, 200], {n: 1})
    expected = l1.delta(l2.as_float())
    assert actual == approx(expected, abs=1e-1)  # at most 0.1 error


def test_kamidelta_2f2():
    x0, x1 = sp.symbols("x:2")
    y0, y1 = sp.symbols("y:2")
    trajectory = {x0: 1, x1: 1, y0: 0, y1: -1}
    start = {x0: 1, x1: 1, y0: -1, y1: -1}
    m = pFq(2, 2, -1).trajectory_matrix(trajectory, start)
    actual = m.kamidelta()[0]
    l1, l2 = m.limit({n: 1}, [100, 200], {n: 1})
    expected = l1.delta(l2.as_float())
    assert actual == approx(expected, abs=1e-1)  # at most 0.1 error


def test_kamidelta_3f2():
    x0, x1, x2 = sp.symbols("x:3")
    y0, y1 = sp.symbols("y:2")
    trajectory = {x0: -1, x1: 1, x2: -2, y0: 2, y1: -3}
    start = trajectory
    m = pFq(3, 2, -1).trajectory_matrix(trajectory, start)
    actual = m.kamidelta(depth=100)[0]
    l1, l2 = m.limit({n: 1}, [100, 200], {n: 1})
    expected = l1.delta(l2.as_float())
    assert actual == approx(expected, abs=1e-1)  # at most 0.1 error



import sys
import warnings
from packaging.version import parse

current = sys.argv[1].lstrip("v")
latest = sys.argv[2].lstrip("v")

curr = parse(current)
last = parse(latest)

if curr < last:
    sys.exit(f"Tag v{current} is older than latest tag v{latest}")

if curr == last:
    warnings.warn(
        f"Tag v{current} is the same as latest tag v{latest}. "
        "Deployment will fail if this version has already been deployed."
    )



from __future__ import annotations

from typing import Callable

import sympy as sp
from flint import fmpq_mat, fmpq

import ramanujantools as rt
from ramanujantools import Position
from ramanujantools.utils import batched, Batchable


class NumericMatrix(fmpq_mat):
    @staticmethod
    def eye(N: int):
        """
        Returns an identity matrix of size N
        """
        retval = NumericMatrix(N, N)
        for i in range(N):
            retval[i, i] = 1
        return retval

    @staticmethod
    def lambda_from_rt(matrix: rt.Matrix) -> Callable:
        """
        Returns a function that evaluates the matrix at given a point
        and returns it as a NumericMatrix.
        """

        def _mul(args, subs):
            result = fmpq(1)
            for arg in args:
                val = arg(subs) if callable(arg) and subs is not None else arg
                result *= val
            return result

        def flint_lambdify(expr):
            """
            A recursive walker that constructs a flint based lambda expression
            from the sympy expression.
            """
            if isinstance(expr, sp.Integer):
                return fmpq(int(expr))
            elif isinstance(expr, sp.Rational):
                return fmpq(expr.p, expr.q)
            elif isinstance(expr, sp.Symbol):
                # Return a function that takes substitutions as a dict
                return lambda subs: fmpq(subs[expr])
            elif isinstance(expr, sp.Add):
                args = [flint_lambdify(arg) for arg in expr.args]
                if any(callable(arg) for arg in args):
                    return lambda subs: sum(
                        arg(subs) if callable(arg) else arg for arg in args
                    )
                else:
                    return sum(args)
            elif isinstance(expr, sp.Mul):
                args = [flint_lambdify(arg) for arg in expr.args]
                if any(callable(arg) for arg in args):
                    return lambda subs: _mul(args, subs)
                else:
                    return _mul(args, None)
            elif isinstance(expr, sp.Pow):
                base = flint_lambdify(expr.base)
                exp = flint_lambdify(expr.exp)
                if callable(base) or callable(exp):
                    return lambda subs: (base(subs) if callable(base) else base) ** int(
                        exp(subs) if callable(exp) else exp
                    )
                else:
                    return base ** int(exp)
            elif isinstance(expr, sp.MatrixBase):
                # Recursively convert each element
                rows, cols = expr.shape
                data = [
                    [flint_lambdify(expr[i, j]) for j in range(cols)]
                    for i in range(rows)
                ]

                return lambda subs: NumericMatrix(
                    [
                        [cell(subs) if callable(cell) else cell for cell in row]
                        for row in data
                    ]
                )

            else:
                raise NotImplementedError(f"Unsupported sympy type: {type(expr)}")

        return flint_lambdify(matrix)

    @staticmethod
    @batched("iterations")
    def walk(
        matrix: rt.Matrix,
        trajectory: Position,
        iterations: Batchable[int],
        start: Position,
    ) -> Batchable[NumericMatrix]:
        results = []
        position = start.copy()
        fast_subs = NumericMatrix.lambda_from_rt(matrix)
        retval = NumericMatrix.eye(matrix.rows)
        for depth in range(0, iterations[-1]):
            if depth in iterations:
                results.append(retval)
            retval *= fast_subs(position)
            position += trajectory
        results.append(retval)  # Last matrix, for iterations[-1]
        return results

    def to_rt(self) -> rt.Matrix:
        return rt.Matrix(self.nrows(), self.ncols(), list(self))

    def __neg__(self) -> NumericMatrix:
        return NumericMatrix(super().__neg__())

    def __add__(self, other) -> NumericMatrix:
        return NumericMatrix(super().__add__(other))

    def __radd__(self, other) -> NumericMatrix:
        return NumericMatrix(super().__radd__(other))

    def __sub__(self, other) -> NumericMatrix:
        return NumericMatrix(super().__sub__(other))

    def __rsub__(self, other) -> NumericMatrix:
        return NumericMatrix(super().__rsub__(other))

    def __mul__(self, other) -> NumericMatrix:
        return NumericMatrix(super().__mul__(other))

    def __rmul__(self, other) -> NumericMatrix:
        return NumericMatrix(super().__rmul__(other))

    def __div__(self, other) -> NumericMatrix:
        return NumericMatrix(super().__div__(other))



import sympy as sp
from sympy.abc import x, y, n

from typing import Collection

from ramanujantools import GenericPolynomial
from ramanujantools.pcf import PCF
from ramanujantools.cmf import FFbar


class FFbarSolver:
    @staticmethod
    def _polynomial_coefficients(
        poly, variables: Collection[sp.Symbol] = (x, y)
    ) -> list:
        r"""
        Returns the coefficients of all monomials of `poly` in `variables`.
        By default assumes variables are $(x, y)$.
        """
        return sp.Poly(poly, *variables).coeffs()

    @staticmethod
    def _solve_equations(expressions: sp.Expr) -> list:
        r"""
        Returns a list of solutions that solve `expressions`.

        The equation system represented by `expressions` is [expr = 0 for expr in expressions].
        """
        return sp.solve(expressions, dict=True, manual=True)

    @staticmethod
    def from_pcf(pcf: PCF) -> list[FFbar]:
        r"""
        Attempts to construct a 2d FFbar CMF,
        such that Mx is the matrix representation of `pcf`,
        and My is as generic as possible.
        """
        a_deg, b_deg = pcf.degrees()
        deg = max(a_deg, (b_deg + 1) // 2)
        f, _ = GenericPolynomial.of_combined_degree(
            deg=deg, var_name="c", variables=[x, y]
        )
        fbar, _ = GenericPolynomial.of_combined_degree(
            deg=deg, var_name="d", variables=[x, y]
        )

        equations = [
            *FFbarSolver._polynomial_coefficients(
                pcf.a_n.subs(n, x) - FFbar.A(f.subs(y, 1), fbar.subs(y, 1))
            ),
            *FFbarSolver._polynomial_coefficients(
                pcf.b_n.subs(n, x) - FFbar.B(f.subs(y, 1), fbar.subs(y, 1))
            ),
        ]

        for solution in FFbarSolver._solve_equations(equations):
            f = f.subs(solution).simplify()
            fbar = fbar.subs(solution).simplify()
            return FFbarSolver.solve_ffbar(f, fbar)

    @staticmethod
    def solve_ffbar(f: sp.Expr, fbar: sp.Expr) -> list[FFbar]:
        r"""
        Returns all pairs of subsets of the input $f(x, y), \bar{f}(x, y)$ such that
        each one is a valid solution of the linear condition and the quadratic condition.
        """
        equations = [
            *FFbarSolver._polynomial_coefficients(FFbar.linear_condition(f, fbar)),
            *FFbarSolver._polynomial_coefficients(FFbar.quadratic_condition(f, fbar)),
        ]
        solutions = FFbarSolver._solve_equations(equations)
        return [
            FFbar(f.subs(solution).simplify(), fbar.subs(solution).simplify())
            for solution in solutions
        ]



from pytest import approx

import sympy as sp
from mpmath import mp

from ramanujantools import Matrix, Limit
from .limit import most_round_in_range


def limit_for_tests(matrix: Matrix) -> Limit:
    return Limit(matrix, matrix)


def test_as_rational():
    p = 2
    q = 3
    limit = limit_for_tests(Matrix([[0, p], [1, q]]))
    assert sp.Rational(p, q) == limit.as_rational()


def test_as_rational_higher_order():
    p = 2
    q = 3
    limit = limit_for_tests(Matrix([[1, 2, p], [3, 4, q], [5, 6, 7]]))
    assert sp.Rational(p, q) == limit.as_rational()


def test_as_float():
    p = 2
    q = 3
    limit = limit_for_tests(Matrix([[0, p], [1, q]]))
    assert p / q == approx(limit.as_float(), 1e-7)


def test_precision_exact():
    a = 5
    b = a - 1
    desired_error = 5
    denominator = 10**desired_error
    limit = Limit(
        Matrix([[a, b], [denominator, denominator]]), Matrix([[0, a], [1, denominator]])
    )
    assert desired_error == limit.precision()


def test_precision_floor():
    a = 5
    b = a - 2
    desired_error = 5
    denominator = 10**desired_error
    limit = Limit(
        Matrix([[a, b], [denominator, denominator]]), Matrix([[0, a], [1, denominator]])
    )
    assert desired_error - 1 == limit.precision()


def test_rounding_in_range():
    for num in [0.49, 0.4999999, 0.5, 0.500001, 0.51]:
        err = 5 * 1e-2
        rounded = eval(most_round_in_range(num, err))

        assert abs(num - rounded) < err


def test_rounding_whole_number():
    num = 0.9999999
    err = 5 * 1e-5
    assert "1" == most_round_in_range(num, err)


def test_rounding_small_change():
    num = mp.mpf(0.9999999)
    err = mp.mpf(5 * 1e-15)
    assert "0.9999999" == most_round_in_range(num, err)


def test_repr():
    limit = limit_for_tests(Matrix([[1, 2], [3, 4]]))
    assert limit == eval(repr(limit))



from .coboundary_solver import CoboundarySolver
from .euler_solver import EulerSolution, EulerSolver
from .ffbar_solver import FFbarSolver

__all__ = ["CoboundarySolver", "EulerSolution", "EulerSolver", "FFbarSolver"]



import sympy as sp
from sympy.abc import n

from ramanujantools import Matrix
from ramanujantools.flint_core import flint_ctx, SymbolicMatrix


def flintify(matrix: Matrix, fmpz=True) -> SymbolicMatrix:
    ctx = flint_ctx(matrix.free_symbols, fmpz)
    return SymbolicMatrix.from_sympy(matrix, ctx)


def test_factor():
    matrix = Matrix(
        [
            [1, n**2 + n, n**2 - n + 5],
            [3 * n + 9, n**2 - 1, 1 / (n + 1)],
            [n**2 + 7, n - 2, (n + 1) * (n - 3)],
        ],
    )

    assert matrix.applyfunc(sp.factor) == flintify(matrix).factor()


def test_mul():
    m1 = Matrix([[0, n**2], [1, 1 / n]])
    m2 = Matrix([[3, n - 2], [n, 5]])

    assert flintify(m1 * m2) == flintify(m1) * flintify(m2)
    assert flintify(m2 * m1) == flintify(m2) * flintify(m1)


def test_mul_scalar():
    m = Matrix([[0, n**2], [1, 1 / n]])

    assert flintify(m * 17) == flintify(m) * 17
    assert flintify(m * 17) == 17 * flintify(m)


def test_walk():
    matrix = Matrix(
        [
            [1, n**2 + n, n**2 - n + 5],
            [3 * n + 9, n**2 - 1, 1 / (n + 1)],
            [n**2 + 7, n - 2, (n + 1) * (n - 3)],
        ],
    )

    expected = (matrix * matrix.subs({n: n + 1}) * matrix.subs({n: n + 2})).factor()
    assert expected == flintify(matrix).walk({n: 1}, 3, {n: n}).factor()


def test_walk_rational():
    matrix = Matrix(
        [
            [1, n**2 + n, n**2 - n + 5],
            [3 * n + 9, n**2 - 1, 1 / (n + 1)],
            [n**2 + 7, n - 2, (n + 1) * (n - 3)],
        ],
    )

    expected = (
        matrix.subs({n: n / 2})
        * matrix.subs({n: n / 2 + 1})
        * matrix.subs({n: n / 2 + 2})
    ).factor()

    assert expected == flintify(matrix, fmpz=False).walk({n: 1}, 3, {n: n / 2}).factor()



from .pcf import PCF
from .hypergeometric import (
    HypergeometricLimit,
    Hypergeometric1F1Limit,
    Hypergeometric2F1Limit,
)

__all__ = [
    "PCF",
    "HypergeometricLimit",
    "Hypergeometric1F1Limit",
    "Hypergeometric2F1Limit",
]



import pytest

import sympy
from sympy import Poly
from sympy.abc import x

from ramanujantools.solvers import EulerSolver, EulerSolution

"""
Looking for Euler solutions of the form
        $$b(x) = -h_1(x) h_1(x)$$
        $$f(x)a(x) = f(x-1)h_1(x) + f(x+1)h_2(x+1)$$

Either there is a trivial solution (f=1), general solution (general f), or no solution.
Right now, the EulerSolver only works when all the polynomials are over Z.
"""

sqrt5 = sympy.sqrt(5)


@pytest.mark.parametrize(
    "h1,h2",
    [
        # degree(h_1) = degree(h_2) = degree(a):=degree(h_1+h_2)
        (1, 1),
        (x**3, x**3),
        (x * x * (x + 5), x * (x - 3) * (x - 10)),
        # degree(h_1) = degree(h_2) > degree(a):=degree(h_1+h_2)
        (x**3, -(x**3)),
        (-(x**3), x**3),
        ((x + 1) * (x + 2), -(x + 1) * (x + 4)),
        # degree(h_1) < degree(h_2)
        (1, x),
        (1, x**2),
        (x + 1, x**2),
        ((x + 1), (x - 1) * (x - 2)),
        # degree(h_1) > degree(h_2)
        (x, 1),
        (x**2, 1),
        (x**2, x + 1),
        ((x + 1) * (x - 5) * (x + 9), (x - 1) * (x - 2)),
        # some non rational algebraic examples:
        ((1 + sqrt5) / 2, (1 - sqrt5) / 2),
        ((1 + sqrt5) / 2, x * (1 - sqrt5) / 2),
        (-x * (1 + sqrt5) / 2, x * (1 - sqrt5) / 2),
        (x**2 * (1 + sqrt5), x**2 * (1 - sqrt5)),
    ],
)
def test_trivial_euler_family(h1: Poly, h2: Poly):
    h2 = Poly(h2, x)  # needed for the subs(x, x+1) called below
    original_solution = EulerSolution(h1, h2, h1 + h2.subs(x, x + 1), -h1 * h2, 1)
    solutions = EulerSolver.solve_for(original_solution.a, original_solution.b)
    assert original_solution in solutions


@pytest.mark.parametrize(
    "h1,h2,f",
    [
        (x**4, x**4, 2 * x + 1),
        (x**5, x**5, 3 * x**2 + 3 * x + 1),
        (x**5, x**5, x**2 + x + 1),
        (x**5, x**5, 5 * x**4 + 10 * x**3 + 19 * x**2 + 14 * x + 4),
        (x**7, x**7, 2 * x**2 + 2 * x + 1),
    ],
)
def test_non_trivial_euler_family(h1: Poly, h2: Poly, f: Poly):
    h1 = Poly(h1, x)
    h2 = Poly(h2, x)
    b = Poly(-h1 * h2, x)

    f = Poly(f, x)
    g = Poly(f.subs(x, x - 1) * h1 + f.subs(x, x + 1) * h2.subs(x, x + 1), x)
    a, r = sympy.div(g, f)
    assert r == 0  # make sure that this is a valid solution

    original_solution = EulerSolution(h1, h2, a, b, f)
    solutions = EulerSolver.solve_for(original_solution.a, original_solution.b)
    assert original_solution in solutions


@pytest.mark.parametrize("a,b", [(3 * x + 1, x), (2 * x + 3, x), (x**2 + x**3, x**4)])
def test_no_euler_solution(a: Poly, b: Poly):
    assert len(EulerSolver.solve_for(a, b)) == 0


def test_solution_equality():
    # Two Euler solutions should be the same if they have the same h_1,h_2,a,b , and their f polynomials are
    # the same up to a scalar multiplication
    x_poly = Poly(x)
    sol1 = EulerSolution(h_1=x_poly, h_2=x_poly, a=x_poly, b=x_poly, f=x_poly)
    sol2 = EulerSolution(h_1=x_poly, h_2=x_poly, a=x_poly, b=x_poly, f=x_poly)
    assert sol1 == sol2

    sol1 = EulerSolution(h_1=x_poly, h_2=x_poly, a=x_poly, b=x_poly, f=-3 * x_poly)
    sol2 = EulerSolution(h_1=x_poly, h_2=x_poly, a=x_poly, b=x_poly, f=8 * x_poly)
    assert sol1 == sol2

    sol1 = EulerSolution(h_1=x_poly, h_2=x_poly, a=x_poly, b=x_poly, f=x_poly)
    sol2 = EulerSolution(h_1=x_poly, h_2=x_poly, a=2 * x_poly, b=x_poly, f=x_poly)
    assert sol1 != sol2

    sol1 = EulerSolution(h_1=x_poly, h_2=x_poly, a=x_poly, b=x_poly, f=2 * x_poly)
    sol2 = EulerSolution(h_1=x_poly, h_2=x_poly, a=2 * x_poly, b=x_poly, f=x_poly)
    assert sol1 != sol2


def test_non_unique_solution():
    # Tests a case where f is generic, and doesn't have any restrictions
    f_0 = sympy.Symbol("f_0")
    original_solution = EulerSolution(
        h_1=Poly(1, x), h_2=Poly(1, x), a=Poly(2, x), b=Poly(-1, x), f=Poly(x + f_0)
    )
    solution = EulerSolver.solve_for_decomposition_with_degree(
        a=original_solution.a,
        h_1=original_solution.h_1,
        h_2=original_solution.h_2,
        d_f=1,
    )
    assert original_solution == solution



import pytest
import sympy
from sympy.abc import x, y, z
from ramanujantools.generic_polynomial import GenericPolynomial


@pytest.mark.parametrize(
    "n, num_var, output",
    [
        (5, 1, [[5]]),
        (5, 2, [[0, 5], [1, 4], [2, 3], [3, 2], [4, 1], [5, 0]]),
        (
            5,
            3,
            [
                [0, 0, 5],
                [0, 1, 4],
                [0, 2, 3],
                [0, 3, 2],
                [0, 4, 1],
                [0, 5, 0],
                [1, 0, 4],
                [1, 1, 3],
                [1, 2, 2],
                [1, 3, 1],
                [1, 4, 0],
                [2, 0, 3],
                [2, 1, 2],
                [2, 2, 1],
                [2, 3, 0],
                [3, 0, 2],
                [3, 1, 1],
                [3, 2, 0],
                [4, 0, 1],
                [4, 1, 0],
                [5, 0, 0],
            ],
        ),
    ],
)
def test_sum_to(n: int, num_var: int, output: list[list[int]]):
    for elements in GenericPolynomial._sum_to(n=n, num_var=num_var):
        assert elements in output
        output.remove(elements)

    assert len(output) == 0


def test_generic_of_degree():
    f, _ = GenericPolynomial.of_degree(deg=3, var_name="f_", s=x, monic=False)
    f_0, f_1, f_2, f_3 = sympy.symbols("f_0 f_1 f_2 f_3")
    assert f == f_0 + f_1 * x + f_2 * x**2 + f_3 * x**3

    f, _ = GenericPolynomial.of_degree(deg=3, var_name="f_", s=x, monic=True)
    assert f == f_0 + f_1 * x + f_2 * x**2 + x**3


def test_generic_of_combined_degree():
    f, _ = GenericPolynomial.of_combined_degree(deg=2, var_name="f_", variables=[x, y])
    # sympy.symbols consider comma (,) as a separator, so need to use sympy.Symbol instead
    f_00 = sympy.Symbol("f_(0,0)")
    f_10 = sympy.Symbol("f_(1,0)")
    f_01 = sympy.Symbol("f_(0,1)")
    f_20 = sympy.Symbol("f_(2,0)")
    f_11 = sympy.Symbol("f_(1,1)")
    f_02 = sympy.Symbol("f_(0,2)")
    assert f == f_00 + f_10 * x + f_01 * y + f_20 * x**2 + f_11 * x * y + f_02 * y**2


def test_symmetric_poly():
    symm0, symm1, symm2, symm3 = GenericPolynomial.symmetric_polynomials(x, y, z)
    assert symm0 == 1
    assert symm1 == x + y + z
    assert symm2 == x * y + y * z + z * x
    assert symm3 == x * y * z


def test_as_symmetric():
    x, y, z, w = sympy.symbols("x y z w")
    s1, s2, s3, s4 = sympy.symbols("s_1 s_2 s_3 s_4")

    # two symmetric variables
    p = (x + y) * (x * y) * z + z**2 * (x**2 + y**2) + (x + y) + 1
    symm_p = s1 * s2 * z + z**2 * (s1**2 - 2 * s2) + s1 + 1
    result = GenericPolynomial.as_symmetric(
        polynomial=p, symm_symbols=[x, y], symm_var_name="s_"
    )
    assert sympy.simplify(result - symm_p) == 0

    # three symmetric variables
    p = (x + y + z) * (x * y * z) * w + w**2 * (x**2 + y**2 + z**2) + (x + y + z) + 1
    symm_p = s1 * s3 * w + w**2 * (s1**2 - 2 * s2) + s1 + 1
    result = GenericPolynomial.as_symmetric(
        polynomial=p, symm_symbols=[x, y, z], symm_var_name="s_"
    )
    assert sympy.simplify(result - symm_p) == 0



from pytest import approx

import mpmath as mp
from sympy.abc import c, n

from ramanujantools import Matrix, LinearRecurrence
from ramanujantools.pcf import PCF
from ramanujantools.pcf.pcf import is_deflatable


def test_repr():
    pcf = PCF(1 + n, 3 - n)
    assert pcf == eval(repr(pcf))


def test_degrees():
    pcf = PCF(1 + n - n**2, 3 - n**9)
    assert (2, 9) == pcf.degrees()


def test_singular_points():
    a_n = n + 1
    b_n = (n + 1) * (n - 17) * (n + 59) * (n - 102)
    pcf = PCF(a_n, b_n)
    assert pcf.singular_points() == [{n: 17}, {n: 102}]


def test_limit_as_float():
    pcf = PCF(5 + 10 * n, 1 - 9 * n**2)
    expected = (4 ** (1 / 3) + 1) / (4 ** (1 / 3) - 1)
    assert expected == approx(pcf.limit(100).as_float(), 1e-4)


def test_inflate_constant():
    c = 3
    a_n = n + 4
    b_n = n**2
    pcf = PCF(a_n, b_n)
    assert PCF(c * a_n, c**2 * b_n) == pcf.inflate(c)


def test_inflate_symbol():
    a_n = n + 4
    b_n = n**2
    pcf = PCF(a_n, b_n)
    assert PCF(c * a_n, c**2 * b_n) == pcf.inflate(c)


def test_inflate_poly():
    a_n = n + 4
    b_n = n**2
    c_n = n**7 + 5 * n - 3
    pcf = PCF(a_n, b_n)
    assert PCF(c_n * a_n, c_n.subs({n: n - 1}) * c_n * b_n) == pcf.inflate(c_n)


def test_deflate_constant():
    a_n = n + 4
    b_n = n**2
    pcf = PCF(c * a_n, c**2 * b_n)
    assert PCF(a_n, b_n) == pcf.deflate(c)


def test_deflate_poly():
    a_n = n + 4
    b_n = n**2
    c_n = n**7 + 5 * n - 3
    pcf = PCF(c_n * a_n, c_n.subs({n: n - 1}) * c_n * b_n)
    assert PCF(a_n, b_n) == pcf.deflate(c_n)


def test_is_deflatable():
    a_factors = {n**2 + 1: 1}
    b_factors = {n**2 - 2 * n + 2: 1, n**2 + 1: 1}
    assert is_deflatable(a_factors, b_factors, n**2 + 1)


def test_deflate_all():
    c_n = c**2 * (7 * n - 13 * c) * (2 * n - 5) ** 4 * (3 * n + 11)
    a_n = n + c
    b_n = n**2 - c * n
    pcf1 = PCF(c_n * a_n, c_n.subs({n: n - 1}) * c_n * b_n)
    pcf2 = PCF(n**2 + 1, n**4 - 2 * n**3 + 3 * n**2 - 2 * n + 2)
    pcf3 = PCF(
        9 * n**4 + 72 * n**3 + 201 * n**2 + 228 * n + 85,
        -18 * n**8
        - 225 * n**7
        - 1131 * n**6
        - 2904 * n**5
        - 3932 * n**4
        - 2400 * n**3
        + 66 * n**2
        + 769 * n
        + 255,
    )
    assert PCF(a_n, b_n) == pcf1.deflate_all()
    assert PCF(1, 1) == pcf2.deflate_all()
    assert (
        PCF(3 * n**2 + 9 * n + 5, -2 * n**4 - 9 * n**3 - 9 * n**2 + n + 3)
        == pcf3.deflate_all()
    )


def test_pcf_construction_from_matrix():
    matrix = Matrix(
        [
            [n * (c * n + 1), n * (2 * n + 1) * (c * n + 1)],
            [2 * n, n * (c * n + 4 * n + 1)],
        ]
    )
    assert PCF(
        (c + 2) * (n + 1) * (2 * n + 1), -n * (n + 1) * (c * n - 1) * (c * n + 1)
    ) == PCF(matrix)


def test_pcf_construction_from_linear_recurrence():
    a_n = 2 * n**2 + 3 * n
    b_n = 5 * n**2 + 7 * n
    recurrence = LinearRecurrence([-1, a_n, b_n])
    assert PCF(a_n, b_n) == PCF(recurrence)


def test_blind_delta():
    pcf = PCF(34 * n**3 + 51 * n**2 + 27 * n + 5, -(n**6))
    depth = 2000
    delta = pcf.delta(depth)
    assert delta > 0.08


def test_precision_e():
    pcf = PCF(n, n)
    assert pcf.limit(2**10 + 1).precision() == 2645


def test_precision_phi():
    pcf = PCF(1, 1)
    assert pcf.limit(2**10 + 1).precision() == 427


def test_delta_sequence_agrees_with_delta():
    pcf = PCF(2 * n + 1, n**2)
    depth = 50
    limit = 4 / mp.pi

    actual_deltas = pcf.delta_sequence(depth, limit)
    expected_deltas = []
    for dep in range(1, depth):
        expected_deltas.append(pcf.delta(dep, limit))

    assert expected_deltas == actual_deltas


def test_blind_delta_sequence_agrees_with_blind_delta():
    pcf = PCF(2 * n + 1, n**2)
    depth = 50
    L = pcf.limit(2 * depth).as_float()

    expected_deltas = [pcf.delta(d, L) for d in range(1, depth)]
    actual_deltas = pcf.delta_sequence(depth)

    assert expected_deltas == actual_deltas



import sympy as sp
from sympy.abc import n

from ramanujantools.pcf import PCF


def p2m(expression: sp.Expr) -> str:
    """Function converts syntax from python to mathematica"""
    return str(expression).replace("(", "[").replace(")", "]").replace("sqrt", "Sqrt")


def HypergeometricLimit(pcf: PCF):
    """Factory function that creates an instance of a hpyergeometric limit class (either 1f1 or 2f1)."""
    supported_degrees = {
        (1, 1): Hypergeometric1F1Limit,
        (1, 2): Hypergeometric2F1Limit,
    }
    if pcf.degrees() not in supported_degrees:
        raise ValueError(
            (
                "Attempted to evaluate hypergeometric limit of {} of degrees {}. "
                "Supported degrees are {}"
            ).format(pcf, pcf.degrees(), supported_degrees.keys())
        )
    return supported_degrees[pcf.degrees()](pcf)


class HypLimitInterface:
    def __init__(self, pcf: PCF):
        pass

    def __eq__(self, other):
        pass

    def limit(self):
        """Attempts to calculate the limit using sympy's hypergeometric functions"""
        pass

    def as_mathematica_prompt(self):
        """Returns a mathematica prompt that will calculate the limit of this pcf"""
        pass

    def subs(self, *args, **kwargs):
        """Substitutes all parameters"""
        pass


class Hypergeometric1F1Limit(HypLimitInterface):
    def __init__(self, pcf: PCF):
        a, c = sp.Poly(pcf.a_n, n).all_coeffs()
        b, d = sp.Poly(pcf.b_n, n).all_coeffs()

        self.alpha = d / b
        self.beta = (c * a + b) / a**2
        self.z = b / a**2
        self.a = a

    def __eq__(self, other):
        return (
            self.alpha == other.alpha and self.beta == other.beta and self.z == other.z
        )

    def limit(self):
        """Attempts to calculate the limit using sympy's 1f1 functions"""
        return sp.simplify(
            sp.hyperexpand(
                self.a
                * self.beta
                * (
                    sp.hyper([self.alpha], [self.beta], self.z)
                    / sp.hyper([self.alpha + 1], [self.beta + 1], self.z)
                )
            )
        )

    def as_mathematica_prompt(self):
        """Returns a mathematica prompt that will calculate the limit of this pcf"""
        return (
            f"({p2m(self.a)}) * ({p2m(self.beta)}) * "
            f"Hypergeometric1F1[{p2m(self.alpha)}, {p2m(self.beta)}, {p2m(self.z)}] / "
            f"Hypergeometric1F1[{p2m(self.alpha + 1)}, {p2m(self.beta + 1)}, {p2m(self.z)}]"
        )

    def subs(self, *args, **kwargs):
        """Substitutes all parameters"""
        import copy

        retval = copy.deepcopy(self)
        retval.alpha = retval.alpha.subs(*args, **kwargs)
        retval.beta = retval.beta.subs(*args, **kwargs)
        retval.z = retval.z.subs(*args, **kwargs)
        return retval


class Hypergeometric2F1Limit(HypLimitInterface):
    def __init__(self, pcf: PCF):
        def solve(expression):
            """Returns a list of the roots of `expression` with multiplicites"""
            return sum(
                [
                    ([sp.simplify(root)] * multiplicity)
                    for root, multiplicity in sp.roots(expression, n).items()
                ],
                [],
            )

        e, d = sp.Poly(pcf.a_n, n).all_coeffs()
        c, b, a = sp.Poly(pcf.b_n, n).all_coeffs()
        delta = e**2 + 4 * c
        self.sqrt_delta = sp.simplify(sp.sign(e) * sp.root(delta, 2))
        self.alpha1, self.alpha2 = solve(c * n**2 - b * n + a)
        self.z = sp.simplify((1 - e / self.sqrt_delta) / 2)
        self.beta = sp.simplify(((b + c) * self.z) / c + d / self.sqrt_delta)

    def __eq__(self, other):
        return (
            self.alpha1 == other.alpha1
            and self.alpha2 == other.alpha2
            and self.beta == other.beta
            and self.z == other.z
            and self.sqrt_delta == other.sqrt_delta
        )

    def limit(self):
        """Attempts to calculate the limit using sympy's 2f1 functions"""
        return sp.simplify(
            sp.hyperexpand(
                self.sqrt_delta
                * self.beta
                * (
                    sp.hyper([self.alpha1, self.alpha2], [self.beta], self.z)
                    / sp.hyper(
                        [self.alpha1 + 1, self.alpha2 + 1], [self.beta + 1], self.z
                    )
                )
            )
        )

    def as_mathematica_prompt(self):
        """Returns a mathematica prompt that will calculate the limit of this pcf"""
        return (
            f"({p2m(self.sqrt_delta)}) * ({p2m(self.beta)}) * "
            f"Hypergeometric2F1[{p2m(self.alpha1)}, {p2m(self.alpha2)}, {p2m(self.beta)}, {p2m(self.z)}] / "
            f"Hypergeometric2F1[{p2m(self.alpha1 + 1)}, {p2m(self.alpha2 + 1)}, {p2m(self.beta + 1)}, {p2m(self.z)}]"
        )

    def subs(self, *args, **kwargs):
        """Substitutes all parameters"""
        import copy

        retval = copy.deepcopy(self)
        retval.alpha1 = retval.alpha1.subs(*args, **kwargs)
        retval.alpha2 = retval.alpha2.subs(*args, **kwargs)
        retval.beta = retval.beta.subs(*args, **kwargs)
        retval.z = retval.z.subs(*args, **kwargs)
        retval.sqrt_delta = retval.sqrt_delta.subs(*args, **kwargs)
        return retval



from __future__ import annotations

import sympy as sp

import ramanujantools as rt
from ramanujantools import Position
from ramanujantools.flint_core import FlintRational, FlintContext
from ramanujantools.utils import batched, Batchable


class SymbolicMatrix:
    """
    Represents a Matrix of FlintRationals.

    It's logic is limited compared to the main Matrix, as it's designed for bottlenecks.
    """

    def __init__(
        self, rows: int, cols: int, values: list[FlintRational], ctx: FlintContext
    ) -> SymbolicMatrix:
        self._rows = rows
        self._cols = cols
        self.values = values
        self.ctx = ctx

    @staticmethod
    def from_sympy(matrix: rt.Matrix, ctx: FlintContext) -> SymbolicMatrix:
        """
        Converts a Matrix to SymbolicMatrix.
        Args:
            matrix: The matrix as ramanujantools.Matrix
            ctx: The desired mpoly context (which also defines the supported variables)
        """
        values = [FlintRational.from_sympy(cell, ctx) for cell in matrix]
        return SymbolicMatrix(matrix.rows, matrix.cols, values, ctx)

    @staticmethod
    def eye(N: int, ctx: FlintContext) -> SymbolicMatrix:
        """
        Creates an identity matrix of size N.

        Args:
            N: The squared matrix dimension
            ctx: The desired mpoly context (which also defines the supported variables)
        """
        values = [FlintRational.from_sympy(sp.simplify(0), ctx)] * N**2
        current = 0
        while current < N**2:
            values[current] += 1
            current += N + 1
        return SymbolicMatrix(N, N, values, ctx)

    def __getitem__(self, key):
        """
        Returns an element of the matrix.
        Supports both matrix[row, col] and matrix[index] syntax
        """
        if isinstance(key, tuple):
            row = key[0]
            col = key[1]
            return self.values[row * self.cols() + col]
        else:
            return self.values[key]

    def __setitem__(self, key, value):
        """
        Returns an element of the matrix.
        Supports both matrix[row, col] and matrix[index] syntax
        """
        if isinstance(key, tuple):
            row = key[0]
            col = key[1]
            self.values[row * self.cols() + col] = value
        else:
            self.values[key] = value

    def __eq__(self, other: FlintRational):
        return self.values == other.values

    def rows(self):
        return self._rows

    def cols(self):
        return self._cols

    def shape(self):
        return (self.rows(), self.cols())

    def row(self, index: int) -> list[FlintRational]:
        row = []
        for i in range(self.cols()):
            row.append(self[index, i])
        return row

    def col(self, index: int) -> list[FlintRational]:
        col = []
        for i in range(self.rows()):
            col.append(self[i, index])
        return col

    def data(self) -> str:
        data = []
        for row_index in range(self.rows()):
            data.append(self.row(row_index))
        return data

    def __repr__(self) -> str:
        return f"SymbolicMatrix({self.data()})"

    def __str__(self) -> str:
        return f"SymbolicMatrix({self.data()})"

    def __mul__(self, other: SymbolicMatrix | int) -> SymbolicMatrix:
        """
        Multiplies self by another SymbolicMatrix or a scalar.
        """
        if isinstance(other, SymbolicMatrix):
            if self.cols() != other.rows():
                raise ValueError(
                    "Attempting to multiply matrices with incompatible shapes!"
                    f"self.cols()={self.cols()}, other.rows()={other.rows()}"
                )
            elements = []
            for row in range(self.rows()):
                for col in range(other.cols()):
                    current = 0
                    for k in range(self.cols()):
                        current += self[row, k] * other[k, col]
                    elements.append(current)
            return SymbolicMatrix(self.rows(), other.cols(), elements, self.ctx)

        else:
            return SymbolicMatrix(
                self.rows(),
                self.cols(),
                [value * other for value in self.values],
                self.ctx,
            )

    def __rmul__(self, other: SymbolicMatrix | int) -> SymbolicMatrix:
        """
        Multiplies a SymbolicMatrix or a scalar by self.
        """
        if isinstance(other, int):
            return self * other
        return other * self

    def __truediv__(self, other: int) -> SymbolicMatrix:
        """
        Divides self by a scalar
        """
        if isinstance(other, SymbolicMatrix):
            raise ValueError("Attempted to divide by matrix!")
        return SymbolicMatrix(
            self.rows(),
            self.cols(),
            [value / other for value in self.values],
            self.ctx,
        )

    def subs(self, substitutions: dict) -> SymbolicMatrix:
        """
        Substitutes symbols in the matrix.
        """
        return SymbolicMatrix(
            self.rows(),
            self.cols(),
            [value.subs(substitutions) for value in self.values],
            self.ctx,
        )

    def factor(self) -> rt.Matrix:
        """
        Factors all elements in the matrix.
        """
        values = [value.factor() for value in self.values]
        return rt.Matrix(self.rows(), self.cols(), values)

    @batched("iterations")
    def walk(
        self, trajectory: dict, iterations: Batchable[int], start: dict
    ) -> Batchable[SymbolicMatrix]:
        r"""
        Returns the multiplication result of walking in a certain trajectory.

        The `walk` operation is defined as $\prod_{i=0}^{n-1}M(s_0 + i \cdot t_0, ..., s_k + i \cdot t_k)$,
        where `M=self`, `(t_0, ..., t_k)=trajectory`, `n=iterations` and `(s_0, ..., s_k)=start`.

        This is a generalization of the basic (and most common) case $\prod_{i=0}^{n-1}M(s+i)$,
        where M=self, n=iterations and s=start.

        Args:
            trajectory: the trajectory of a single step in the walk, as defined above.
            iterations: The amount of multiplications to perform. Can be an integer value or a list of values.
            start: the starting point of the matrix multiplication
        Returns:
            The walk multiplication matrix as defined above.
            If iterations is list, returns a list of matrices.
        Raises:
            ValueError: If `self` is not a square matrix,
                        if `start` and `trajectory` have different keys,
                        if `iterations` contains duplicate values
        """
        position = Position(start)
        trajectory = Position(trajectory)
        results = []
        matrix = SymbolicMatrix.eye(self.rows(), self.ctx)
        for depth in range(0, iterations[-1]):
            if depth in iterations:
                results.append(matrix)
            matrix *= self.subs(position)
            position += trajectory
        results.append(matrix)  # Last matrix, for iterations[-1]
        return results



from sympy.abc import n

from ramanujantools import Matrix, Position
from ramanujantools.flint_core import NumericMatrix


def test_conversion():
    m = Matrix(
        [
            [(n - 1) * (n**2 - n + 1) / n**3, -1 / n**3],
            [1 / n**3, (n + 1) * (n**2 + n + 1) / n**3],
        ]
    )

    for i in range(1, 10):
        assert m.subs({n: i}) == NumericMatrix.lambda_from_rt(m)({n: i}).to_rt()


def test_walk():
    m = Matrix(
        [
            [(n - 1) * (n**2 - n + 1) / n**3, -1 / n**3],
            [1 / n**3, (n + 1) * (n**2 + n + 1) / n**3],
        ]
    )

    assert (
        m.walk({n: 2}, 100, {n: 3})
        == NumericMatrix.walk(m, Position({n: 2}), 100, Position({n: 3})).to_rt()
    )


def test_walk_list():
    m = Matrix(
        [
            [(n - 1) * (n**2 - n + 1) / n**3, -1 / n**3],
            [1 / n**3, (n + 1) * (n**2 + n + 1) / n**3],
        ]
    )

    rt_walk_matrices = m.walk({n: 2}, list(range(1, 100)), {n: 3})
    numeric_walk_matrices = NumericMatrix.walk(
        m, Position({n: 2}), list(range(1, 100)), Position({n: 3})
    )
    for rt_walk_matrix, numeric_walk_matrix in zip(
        rt_walk_matrices, numeric_walk_matrices
    ):
        assert rt_walk_matrix == numeric_walk_matrix.to_rt()



from __future__ import annotations

import mpmath as mp
import sympy as sp
from sympy.abc import n

from ramanujantools import Matrix, Limit, LinearRecurrence
from ramanujantools.utils import batched, Batchable


def is_deflatable(a_factors, b_factors, factor):
    if n in factor.free_symbols:
        return (
            a_factors.get(factor, 0) > 0
            and b_factors.get(factor, 0) > 0
            and b_factors.get(factor.subs({n: n - 1}).expand(), 0) > 0
        )
    else:
        return a_factors.get(factor, 0) > 0 and b_factors.get(factor, 0) > 1


def remove_factor(a_factors, b_factors, factor):
    a_factors[factor] -= 1
    b_factors[factor] -= 1
    b_factors[factor.subs({n: n - 1}).expand()] -= 1


def deflate_constant(a_constant, b_constant):
    factors = sp.factorint(sp.gcd(a_constant**2, b_constant))
    constant = 1
    for root, mul in factors.items():
        constant *= root ** (mul // 2)
    return constant


def content(a, b, variables):
    if len(a.free_symbols | b.free_symbols) == 0:
        return deflate_constant(a, b)

    def factor_list(poly, variables):
        content, factors = sp.factor_list(poly, *variables)
        return content, {
            factor.expand(): power for factor, power in dict(factors).items()
        }

    (a_content, a_factors), (b_content, b_factors) = map(
        lambda p: factor_list(p, variables), [a, b]
    )

    c_n = content(a_content, b_content, [])
    for factor in a_factors:
        while is_deflatable(a_factors, b_factors, factor):
            remove_factor(a_factors, b_factors, factor)
            c_n *= factor

    return sp.simplify(c_n)


class PCF:
    """
    Represents a Polynomial Continued Fraction (PCF).
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes a PCF. Accepts one of:
        - a_n: sp.Expr, b_n: sp.Expr
        - recurrence: LinearRecurrence
        - matrix: Matrix.
        """
        if len(args) == 2 and all(isinstance(arg, (int, sp.Expr)) for arg in args):
            a_n, b_n = args
            if b_n == 0:
                raise ValueError("b_n cannot be zero in a PCF!")
            self.recurrence = LinearRecurrence([-1, a_n, b_n])
        elif len(args) == 1 and isinstance(recurrence := args[0], LinearRecurrence):
            self.recurrence = recurrence.normalize()
        elif len(args) == 1 and isinstance(matrix := args[0], Matrix):
            if not matrix.is_square() or matrix.rows != 2:
                raise ValueError(
                    "Only a 2x2 matrix can be converted to a PCF, got {obj.rows}x{obj.cols}"
                )
            self.recurrence = LinearRecurrence(matrix).normalize()
        else:
            raise ValueError(
                f"Invalid PCF arguments: got {', '.join(type(arg).__name__ for arg in args)}"
            )

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"PCF({self.a_n}, {self.b_n})"

    def __eq__(self, other: PCF) -> bool:
        return self.recurrence == other.recurrence

    @property
    def a_n(self) -> sp.Expr:
        return self.recurrence.relation[1]

    @property
    def b_n(self) -> sp.Expr:
        return self.recurrence.relation[2]

    def _latex(self, printer) -> str:
        DEPTH = 3

        def recurse(i: int, depth: int) -> str:
            if i == depth:
                b_n_latex = printer.doprint(self.b_n)
                a_n_latex = printer.doprint(self.a_n)
                return rf"\ddots + \cfrac{{{b_n_latex}}}{{{a_n_latex}}}"
            else:
                a_i = printer.doprint(self.a_n.subs({n: i}))
                b_i = printer.doprint(self.b_n.subs({n: i + 1}))
                return rf"{a_i} + \cfrac{{{b_i}}}{{{recurse(i + 1, DEPTH)}}}"

        a_0 = printer.doprint(self.a_n.subs({n: 0}))
        b_1 = printer.doprint(self.b_n.subs({n: 1}))
        return rf"{a_0} + \cfrac{{{b_1}}}{{{recurse(1, DEPTH)}}}"

    def _repr_latex_(self) -> str:
        return rf"{sp.latex(self)}"

    def degrees(self) -> tuple[int, int]:
        """
        Returns the degrees of a_n and b_n as a tuple: $(deg(a_n), deg(b_n))$
        """
        return tuple(map(lambda p: sp.Poly(p, n).degree(), [self.a_n, self.b_n]))

    def M(self) -> Matrix:
        r"""
        Returns the matrix that represents the PCF recurrence:

        $M = \begin{pmatrix} 0, b_n \cr 1, a_n \end{pmatrix}$
        """
        return self.recurrence.recurrence_matrix

    def A(self) -> Matrix:
        r"""
        Returns the matrix that represents the $a_0$ part:

        $A = \begin{pmatrix} 1, a_0 \cr 0, 1 \end{pmatrix}$
        """
        return Matrix([[1, self.a_n.subs({n: 0})], [0, 1]])

    def inflate(self, c_n: sp.Expr) -> PCF:
        r"""
        Inflates the PCF by a polynomial c.
        See LinearRecurrence.inflate for more details.
        """
        return PCF(self.recurrence.inflate(c_n))

    def deflate(self, c_n: sp.Expr) -> PCF:
        r"""
        Deflates the PCF by a polynomial c.
        See LinearRecurrence.deflate for more details.
        """
        return PCF(self.recurrence.deflate(c_n))

    def deflate_all(self) -> PCF:
        """
        Deflates the PCF to the fullest extent, by calculating the biggest $c_n$ possibly deflated
        """
        return self.deflate(content(self.a_n, self.b_n, [n]))

    def simplify(self) -> PCF:
        """Simplifies the PCF (i.e, simplifies (a, b))"""
        return PCF(self.a_n.factor().simplify(), self.b_n.factor().simplify())

    def subs(self, *args, **kwargs) -> PCF:
        """Substitutes parameters in the PCF"""
        return PCF(self.a_n.subs(*args, **kwargs), self.b_n.subs(*args, **kwargs))

    def singular_points(self) -> list[dict]:
        return [
            solution
            for solution in self.M().singular_points()
            if solution[n].is_integer and solution[n] > 0
        ]

    @batched("iterations")
    def limit(self, iterations: Batchable[int], start: int = 1) -> list[Limit]:
        r"""
        Returns the limit corresponding to calculating the PCF up to a certain depth, including $a_0$

        This is essentially $\A \cdot \prod_{i=0}^{n-1}M(s + i)$ where `n=iterations` and `s=start`.

        Args:
            iterations: The amount of multiplications to perform. Can be an integer value or a list of values.
            start: The n value of the first matrix to be multiplied (1 by default)
        Returns:
            The pcf convergence limit as defined above.
            If iterations is a list, returns a list of limits.
        """
        return self.recurrence.limit(iterations, start=start, initial_values=self.A())

    def delta(self, depth: int, L=None) -> mp.mpf:
        r"""
        Calculates the irrationality measure $\delta$ defined, as:
        $|\frac{p_n}{q_n} - L| = \frac{1}{q_n}^{1+\delta}$

        If the limit is not specified (`L` is None),
        then `L` is approximated as `self.limit(2 * depth).as_float()`.

        Args:
            depth: $n$
            limit: $L$
        Returns:
            the delta value as defined above.
        Raises:
            ValueError: if depth <= 0
        """

        if depth <= 0:
            raise ValueError("Cannot calculate delta up to a non-positive depth")

        if L is None:
            limit_and_L = self.limit([depth, 2 * depth])
            limit, L = limit_and_L[0], limit_and_L[1].as_float()
        else:
            limit = self.limit(depth)
        return limit.delta(L)

    def delta_sequence(self, depth: int, L: float = None) -> list[mp.mpf]:
        r"""
        Calculates the irrationality measure $\delta$ defined, as:
        $|\frac{p_n}{q_n} - L| = \frac{1}{q_n}^{1+\delta}$

        If the limit is not specified (`L` is None),
        then `L` is approximated as `self.limit(2 * depth).as_float()`.

        Args:
            depth: $n$
            limit: $L$
        Returns:
            the delta values for all depths up to `depth` as defined above.
        Raises:
            ValueError: if depth <= 0
        """

        if depth <= 0:
            raise ValueError("Cannot calculate delta up to a non-positive depth")
        iterations = list(range(1, depth))
        if L is None:
            limits_and_L = self.limit(iterations + [depth * 2])
            limits, L = limits_and_L[:-1], limits_and_L[-1].as_float()
        else:
            limits = self.limit(iterations)
        deltas = [limit.delta(L) for limit in limits]
        return deltas

    def kamidelta(self, depth: int = 20) -> mp.mpf:
        r"""
        Uses the Kamidelta alogrithm to predict the delta value of the PCF.
        Effectively calls kamidelta on `M`, the recurrence matrix.

        For more details, see `Matrix.kamidelta`
        """
        return self.M().kamidelta(depth)[0]



import pytest
from sympy.abc import x, y, n

from ramanujantools import GenericPolynomial
from ramanujantools.pcf import PCF
from ramanujantools.cmf.known_cmfs import cmf1
from ramanujantools.solvers import FFbarSolver


@pytest.mark.parametrize("deg", [1, 2])
def test_solver_full_poly(deg: int):
    f, _ = GenericPolynomial.of_combined_degree(deg=deg, var_name="c", variables=[x, y])
    fbar, _ = GenericPolynomial.of_combined_degree(
        deg=deg, var_name="d", variables=[x, y]
    )
    assert len(FFbarSolver.solve_ffbar(f, fbar)) > 0


def test_from_ffbar_gauss():
    pcf = PCF(2 * n + 1, -(n**2))
    solutions = FFbarSolver.from_pcf(pcf)
    assert len(solutions) == 1
    cmf = solutions[0]
    assert pcf.M() == cmf.M(x).subs({x: n})


def test_ffbar_from_cmf1():
    cmf = cmf1()
    pcf = PCF(cmf.trajectory_matrix({x: 1, y: 0}, {x: 1, y: 1}))
    solutions = FFbarSolver.from_pcf(pcf)
    for ffbar in solutions:
        assert (
            len(
                FFbarSolver._solve_equations(
                    [ffbar.M(x) - cmf.M(x), ffbar.M(y) - cmf.M(y)]
                )
            )
            > 0
        )  # Making sure that our CMF is equivalent up to symbol names



def get_parent_module(obj):
    module_tree = getattr(obj, "__module__", None)
    return module_tree.split(".")[0] if module_tree else None


def simplify(obj):
    if get_parent_module(obj) == "ramanujantools":
        return obj.simplify()
    else:
        import sympy as sp

        return sp.simplify(obj)



from __future__ import annotations

from typing import TYPE_CHECKING
from functools import lru_cache, cached_property

import numpy as np
import mpmath as mp
import sympy as sp
from sympy.abc import n

from ramanujantools import Position
from ramanujantools.utils import batched, Batchable
from ramanujantools.flint_core import flint_ctx, SymbolicMatrix, NumericMatrix

if TYPE_CHECKING:
    from ramanujantools import Limit


class Matrix(sp.Matrix):
    """
    Represents a Matrix.

    Inherits from sympy's matrix and supports all of its methods and operators:
    https://docs.sympy.org/latest/modules/matrices/matrices.html
    """

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, *args, **kwargs)

    @staticmethod
    def e(N: int, index: int) -> Matrix:
        """
        Returns a coordinate vector of size N for a given index, i.e,
        a vector of size N of zeroes with 1 in the corresponding index

        Args:
            N: The vector size
            index: The index of the given axis
        Returns:
            The desired coordinate vector described above
        """
        if index >= N:
            raise ValueError(f"Cannot create {index}th axis vector of size {N}")
        return Matrix.eye(N).col(index)

    def __str__(self) -> str:
        return repr(self)

    @classmethod
    def _optimized_eq(cls, a, b) -> bool:
        return (
            a.rows == b.rows
            and a.cols == b.cols
            and all(cell == 0 for cell in (a - b).factor())
        )

    def __eq__(self, other: Matrix) -> bool:
        try:
            return self._optimized_eq(self, other)
        except sp.PolynomialError:
            return sp.Matrix.__eq__(self, other)

    def __hash__(self) -> int:
        return hash(frozenset(self))

    def __call__(self, substitutions: dict) -> Matrix:
        """
        Substitutes symbols in the matrix, in a more math-like syntax.
        """
        return self.subs(substitutions)

    def subs(self, substitutions: Position) -> Matrix:
        """
        Substitutes symbols in the matrix.

        Calls the underlying sympy xreplace method, as subs is too slow and we don't need it's extra functionality
        https://docs.sympy.org/latest/modules/core.html#sympy.core.basic.Basic.xreplace
        """
        return self.xreplace(substitutions)

    def _is_numeric_walk(self, trajectory: dict, start: dict) -> bool:
        trajectory = Position(trajectory)
        start = Position(start)

        subbed_in = trajectory.free_symbols().union(start.free_symbols())
        subbed_out = set(trajectory.keys()).union(set(start.keys()))

        return (self.free_symbols - subbed_out).union(subbed_in) == set()

    def is_square(self) -> int:
        """
        Returns the amount of rows/columns of the square matrix (which is of dimension NxN)
        """
        return self.rows == self.cols

    @cached_property
    def denominator_lcm(self) -> sp.Expr:
        """
        Returns the lcm of all denominators
        """
        divisors = [cell.cancel().as_numer_denom()[1] for cell in self]
        return sp.lcm(divisors)

    def is_polynomial(self) -> bool:
        return self.denominator_lcm == 1

    def as_polynomial(self) -> Matrix:
        """
        Converts the matrix to a polynomial matrix by multiplying it by the denominators lcm.
        """
        return (self * self.denominator_lcm).simplify()

    @cached_property
    def gcd(self) -> sp.Rational:
        """
        Returns the rational gcd of the matrix, which could also be parameteric.
        """
        return sp.gcd(list(self.simplify()))

    def reduce(self) -> Matrix:
        """
        Reduces gcd from the matrix
        """
        # important: must simplify first, both for correctness and performance. reproducible example:
        # t = x*(x - 1)*(x + 1)/(x**2 + x)
        # sp.gcd(t, x) == x
        # sp.gcd(t.simplify(), x) == 1
        return (self / self.gcd).simplify()

    def equal_projectively(self, other: Matrix) -> bool:
        """
        Returns true iff two matrices are equal projectively.
        Two matrices are equal projectively iff self = c * other for some c.
        """
        return self.as_polynomial().reduce() == other.as_polynomial().reduce()

    @lru_cache
    def inverse(self) -> Matrix:
        """
        Inverts the matrix.
        """
        return self.inv()

    @lru_cache
    def simplify(self) -> Matrix:
        """
        Returns a simplified version of matrix
        """
        return Matrix(sp.simplify(self))

    def factor(self) -> Matrix:
        if self.free_symbols:
            return SymbolicMatrix.from_sympy(
                self, flint_ctx(self.free_symbols, fmpz=True)
            ).factor()
        else:
            return self

    def singular_points(self) -> list[dict]:
        r"""
        Calculates the singular points of the matrix,
        i.e, points where $|m| = 0$

        Returns:
            A list of substitution dicts that result in the matrix having a zero determinant.
            That is, for each dict in result, `self.subs(dict).det() == 0`
        """
        return sp.solve(self.det(), dict=True)

    def coboundary(self, U: Matrix, symbol: sp.Symbol = n, sign: bool = True) -> Matrix:
        r"""
        Calculates the coboundary relation of M and U $U^{-1}(n) * M(n) * U^(n+s)$,
        where $M$ is `self` and $s$ is `sign`.

        Args:
            U: The coboundary matrix
            symbol: The symbol to use when calculating the coboundary relation. `n` by default.
            sign: Represents if `symbol` goes up or down in the coboundary relation.
        Returns:
            The coboundary relation as described above
        """
        free_symbols = self.free_symbols.union(U.free_symbols).union({symbol})
        ctx = flint_ctx(free_symbols, fmpz=True)
        return (
            SymbolicMatrix.from_sympy(U.inverse(), ctx)
            * SymbolicMatrix.from_sympy(self, ctx)
            * SymbolicMatrix.from_sympy(
                U.subs({symbol: symbol + (1 if sign else -1)}), ctx
            )
        ).factor()

    def companion_coboundary_matrix(self, symbol: sp.Symbol = n) -> Matrix:
        r"""
        Constructs a new matrix U such that `self.coboundary(U)` is a companion matrix.
        """
        if not (self.is_square()):
            raise ValueError("Only square matrices can have a coboundary relation")
        N = self.rows
        ctx = flint_ctx(self.free_symbols, fmpz=True)
        flint_self = SymbolicMatrix.from_sympy(self, ctx)
        vectors = [SymbolicMatrix.from_sympy(Matrix(N, 1, [1] + (N - 1) * [0]), ctx)]
        for _ in range(1, N):
            vectors.append(flint_self * vectors[-1].subs({symbol: symbol + 1}))
        return Matrix.hstack(*[vector.factor() for vector in vectors]).factor()

    @staticmethod
    def companion_form(values: list[sp.Expr]) -> Matrix:
        N = len(values)
        columns = []
        for i in range(N - 1):
            columns.append(Matrix.e(N, i + 1))
        columns.append(Matrix(list(values)))
        return Matrix.hstack(*columns)

    def is_companion(self) -> bool:
        r"""
        Returns True iff the matrix is a companion matrix.
        """
        if not self.is_square():
            raise ValueError("Attempted to check if a non-square matrix is companion")
        N = self.rows
        for row in range(N):
            for col in range(N - 1):
                if row == col + 1:
                    if self[row, col] != 1:
                        return False
                else:
                    if self[row, col] != 0:
                        return False
        return True

    def as_companion(self, symbol: sp.Symbol = n) -> Matrix:
        r"""
        Converts the matrix to companion form.
        """
        if symbol not in self.free_symbols:
            raise ValueError(
                f"Companionization symbol must be in matrix! matrix={self}, symbol={symbol}"
            )
        U = self.companion_coboundary_matrix(symbol)
        try:
            return self.coboundary(U, symbol)
        except ValueError:
            rank = U.rank()
            symbols = sp.symbols(f"c:{rank}")
            variables = Matrix(symbols + (-1,))
            truncated = U[:, : rank + 1]
            solutions = sp.solve(truncated * variables, symbols)
            elements = [solutions[symbol] for symbol in symbols]
            return Matrix.companion_form(elements)

    @lru_cache
    def _walk_inner(
        self,
        trajectory: Position,
        iterations: tuple[int],
        start: Position,
    ) -> list[Matrix]:
        """
        Internal walk function, used for type conversions and for caching. Do not use directly.
        """
        iterations = list(iterations)
        if self._is_numeric_walk(trajectory, start):
            results = NumericMatrix.walk(self, trajectory, iterations, start)
            return [result.to_rt() for result in results]
        else:
            symbols = self.walk_free_symbols(start)
            as_flint = SymbolicMatrix.from_sympy(
                self, flint_ctx(symbols, fmpz=start.is_polynomial())
            )
            results = as_flint.walk(trajectory, iterations, start)
            return [result.factor() for result in results]

    @batched("iterations")
    def walk(
        self,
        trajectory: dict,
        iterations: Batchable[int],
        start: dict,
    ) -> Batchable[Matrix]:
        r"""
        Returns the multiplication result of walking in a certain trajectory.

        The `walk` operation is defined as $\prod_{i=0}^{n-1}M(s_0 + i \cdot t_0, ..., s_k + i \cdot t_k)$,
        where `M=self`, `(t_0, ..., t_k)=trajectory`, `n=iterations` and `(s_0, ..., s_k)=start`.

        This is a generalization of the basic (and most common) case $\prod_{i=0}^{n-1}M(s+i)$,
        where M=self, n=iterations and s=start.

        Args:
            trajectory: the trajectory of a single step in the walk, as defined above.
            iterations: The amount of multiplications to perform. Can be an integer value or a list of values.
            start: the starting point of the matrix multiplication
        Returns:
            The walk multiplication matrix as defined above.
            If iterations is list, returns a list of matrices.
        Raises:
            ValueError: If `self` is not a square matrix,
                        if `start` and `trajectory` have different keys,
                        if `iterations` contains duplicate values
        """
        if not self.is_square():
            raise ValueError(
                f"Matrix.walk is only supported for square matrices, got a {self.rows}x{self.cols} matrix"
            )

        if start.keys() != trajectory.keys():
            raise ValueError(
                f"`start` and `trajectory` must contain same keys, got "
                f"start={set(start.keys())}, trajectory={set(trajectory.keys())}"
            )

        iterations_set = set(iterations)
        if len(iterations_set) != len(iterations):
            raise ValueError(f"`iterations` values must be unique, got {iterations}")

        if not iterations == sorted(iterations):
            raise ValueError(f"Iterations must be sorted, got {iterations}")

        if not all(depth >= 0 for depth in iterations):
            raise ValueError(
                f"iterations must contain only non-negative values, got {iterations}"
            )
        return self._walk_inner(
            Position(trajectory), tuple(iterations), Position(start)
        )

    def walk_free_symbols(self, start: dict) -> set[sp.Symbol]:
        """
        Returns the expected free_symbols of the expression `self.walk(trajectory, iterations, start)`
        """
        free_symbols = self.free_symbols.copy()
        for key in start.keys():
            free_symbols = free_symbols.union(set(sp.simplify(key).free_symbols))
        for value in start.values():
            free_symbols = free_symbols.union(set(sp.simplify(value).free_symbols))
        return free_symbols

    @batched("iterations")
    def limit(
        self,
        trajectory: dict,
        iterations: Batchable[int],
        start: dict,
        initial_values: Matrix | None = None,
        final_projection: Matrix | None = None,
    ) -> Batchable[Limit]:
        from ramanujantools import Limit

        def walk_function(iterations):
            return self.walk(trajectory, iterations, start)

        return Limit.walk_to_limit(
            iterations, walk_function, initial_values, final_projection
        )

    @staticmethod
    def poincare_poly(poly: sp.PurePoly) -> sp.PurePoly:
        """
        Deflates a polynomial such that all coefficients approach a finite number.
        Assumes polynomial only contain n as a free symbol.
        """
        current_degree = 0
        charpoly_coeffs = poly.all_coeffs()
        for i in range(len(charpoly_coeffs)):
            coeff = charpoly_coeffs[i]
            numerator, denominator = coeff.as_numer_denom()
            degree = sp.Poly(numerator, n).degree() - sp.Poly(denominator, n).degree()
            if (current_degree * i) < degree:
                current_degree = -(degree // -i)  # ceil div trick
        coeffs = [
            (charpoly_coeffs[i] / (n ** (current_degree * i))).limit(n, "oo")
            for i in range(len(charpoly_coeffs))
        ]
        return sp.PurePoly(coeffs, poly.gen)

    def charpoly(self, poincare=False, *args, **kwargs) -> sp.PurePoly:
        """
        Returns the characteristic polynomial of the matrix.

        Calls the underyling sympy.Matrix.charpoly and supports its arguments:
        https://docs.sympy.org/latest/modules/matrices/matrices.html#sympy.matrices.matrixbase.MatrixBase.charpoly

        Args:
            poincare: if True, converts the polynomial to Poincare form. False by default.
        """
        poly = super().charpoly(*args, **kwargs)
        if poincare:
            poly = Matrix.poincare_poly(poly)
        return poly

    def eigenvals(self, poincare=False) -> dict:
        """
        Returns the eigenvalues of the matrix, which are the roots of the characteristic polynomials.

        Args:
            poincare: if True, returns the eigenvalues of the Poincare characteristic polynomial instead.
            False by default.
        """
        charpoly = self.charpoly(poincare)
        return sp.roots(charpoly)

    def sorted_eigenvals(self) -> list:
        """
        Returns the eigenvalues of the matrix in Poincare form, sorted by absolute value in descending order.
        """
        unsorted = self.eigenvals(poincare=True)
        retval = []
        for key, value in unsorted.items():
            retval += [key] * value
        return sorted(
            retval, key=lambda value: abs(value).evalf(chop=True), reverse=True
        )

    def errors(self) -> list:
        """
        Approximate the possible errors of integer recurrence approximations using this Matrix.
        """
        lambdas = [e.evalf(chop=True) for e in self.sorted_eigenvals()]
        deltas = []
        for i in range(1, len(lambdas)):
            deltas.append(sp.log(abs(lambdas[0]) / abs(lambdas[i])))
        return deltas

    def gcd_slope(self, depth=20) -> mp.mpf:
        r"""
        Attempts to perform a linear fit of $\bar{q} = \frac{q_n}{gcd(p_n, q_n)}$ as a function of $n$.

        Args:
            depth: The maximal value of $n$
        Returns:
            The slope of the lienar fit of $\bar{q}$.
        """
        depths = list(range(1, depth))
        q_reduced_list = []
        limits = self.limit({n: 1}, depths, {n: 1})
        for limit in limits:
            q_reduced_list.append(sp.log(limit.as_rational().q).evalf(30))
        fit = np.polyfit(
            np.array(depths), np.array(q_reduced_list, dtype=np.float64), 1
        )
        return mp.mpf(fit[0])

    def kamidelta(self, depth=20) -> list[mp.mpf]:
        r"""
        Predicts the possible delta values of the integer sequence approximation that the matrix generates.

        The irrationality measure $\delta$ is defined as $|\frac{p_n}{q_n} - L| = \frac{1}{\bar{q_n}^{1+\delta}}$.
        As one can tell, the delta is determined by two values:
        The approximation error $L-\frac{p_n}{q_n}$,
        and the gcd of both sequences $gcd(p_n,q_n)$.

        This algorithm approximates both values in order to predict possible delta values.

        Args:
            depth: The maximal depth for the gcd slope fit.
        Returns:
            A list (of size N - 1) containing all predicted delta possibilities.
        """
        errors = self.errors()
        slope = self.gcd_slope(depth)
        return [-1 + error / slope for error in errors]



import sympy as sp

from ramanujantools.cmf import CMF, e, pFq

a0, a1 = sp.symbols("a:2")
b0, b1 = sp.symbols("b:2")
x0, x1, x2, x3 = sp.symbols("x:4")
y0, y1, y2, y3 = sp.symbols("y:4")


def test_trajectory_matrix_simple(benchmark):
    from sympy.abc import x, y

    cmf = e()
    start = {x: 1, y: 1}
    trajectory = {x: 1, y: 1}
    benchmark(CMF.trajectory_matrix, cmf, trajectory, start)


def test_trajectory_matrix_2f2_euler(benchmark):
    x0, x1 = sp.symbols("x:2")
    y0, y1 = sp.symbols("y:2")
    cmf = pFq(2, 2, -1)
    start = {x0: 1, x1: 1, y0: -1, y1: -1}
    trajectory = {x0: 1, x1: 1, y0: 0, y1: -1}
    benchmark(CMF.trajectory_matrix, cmf, trajectory, start)


def test_trajectory_matrix_2f2_deep(benchmark):
    x0, x1 = sp.symbols("x:2")
    y0, y1 = sp.symbols("y:2")
    cmf = pFq(2, 2, -1)
    start = {x0: 1, x1: 1, y0: -1, y1: 1}
    trajectory = {x0: 12, x1: 13, y0: -14, y1: 20}
    benchmark(CMF.trajectory_matrix, cmf, trajectory, start)


def test_trajectory_matrix_3f2(benchmark):
    x0, x1, x2 = sp.symbols("x:3")
    y0, y1 = sp.symbols("y:2")
    cmf = pFq(3, 2, 1)
    start = {x0: 2, x1: 2, x2: 2, y0: 4, y1: 4}
    trajectory = {x0: 5, x1: 5, x2: 5, y0: 10, y1: 10}
    benchmark(CMF.trajectory_matrix, cmf, trajectory, start)


def test_trajectory_matrix_3f2_rational(benchmark):
    x0, x1, x2 = sp.symbols("x:3")
    y0, y1 = sp.symbols("y:2")
    cmf = pFq(3, 2, 1)
    start = {
        x0: sp.Rational(1, 2),
        x1: sp.Rational(1, 2),
        x2: sp.Rational(1, 2),
        y0: sp.Rational(3, 2),
        y1: sp.Rational(3, 2),
    }
    trajectory = {x0: 5, x1: 5, x2: 5, y0: 10, y1: 10}
    benchmark(CMF.trajectory_matrix, cmf, trajectory, start)


def test_trajectory_matrix_3f2_pertubated(benchmark):
    x0, x1, x2 = sp.symbols("x:3")
    y0, y1 = sp.symbols("y:2")
    cmf = pFq(3, 2, 1)
    start = {x0: 2, x1: 2, x2: 2, y0: 4, y1: 4}
    trajectory = {x0: 5, x1: 6, x2: 5, y0: 10, y1: 11}
    benchmark(CMF.trajectory_matrix, cmf, trajectory, start)


def test_trajectory_matrix_4f3(benchmark):
    x0, x1, x2, x3 = sp.symbols("x:4")
    y0, y1, y2 = sp.symbols("y:3")
    cmf = pFq(4, 3, 1)
    start = {x0: 1, x1: 1, x2: 2, x3: 2, y0: 3, y1: 3, y2: 4}
    trajectory = {x0: 1, x1: 1, x2: 2, x3: 2, y0: 3, y1: 3, y2: 4}
    benchmark(CMF.trajectory_matrix, cmf, trajectory, start)


def test_trajectory_matrix_4f3_huge(benchmark):
    x0, x1, x2, x3 = sp.symbols("x:4")
    y0, y1, y2 = sp.symbols("y:3")
    cmf = pFq(4, 3, 1)
    trajectory = {x0: -1, x1: 1, x2: 2, x3: -2, y0: 3, y1: 7, y2: 4}
    start = trajectory
    benchmark(CMF.trajectory_matrix, cmf, trajectory, start)


def test_walk_4f3(benchmark):
    x0, x1, x2, x3 = sp.symbols("x:4")
    y0, y1, y2 = sp.symbols("y:3")
    cmf = pFq(4, 3, 1)
    start = {x0: 1, x1: 1, x2: 2, x3: 2, y0: 3, y1: 3, y2: 4}
    trajectory = {x0: 1, x1: 1, x2: 2, x3: 2, y0: 3, y1: 3, y2: 4}
    benchmark(CMF.walk, cmf, trajectory, 1000, start)


def test_sub_cmf_3f3(benchmark):
    cmf = pFq(3, 3, -1)
    basis = {
        a0: {x0: 0, x1: 1, x2: 0, y0: 0, y1: 1, y2: -1},
        a1: {x0: 1, x1: 0, x2: 1, y0: 0, y1: 0, y2: 0},
        b0: {x0: 0, x1: 0, x2: 0, y0: 0, y1: 1, y2: 1},
        b1: {x0: 1, x1: 0, x2: 0, y0: 1, y1: 0, y2: 0},
    }
    benchmark(CMF.sub_cmf, cmf, basis)



from .cmf import CMF
from .d_finite import DFinite
from .ffbar import FFbar
from .pfq import pFq
from .meijer_g import MeijerG

from .known_cmfs import (
    e,
    pi,
    symmetric_pi,
    zeta3,
    var_root_cmf,
    cmf1,
    cmf2,
    cmf3_1,
    cmf3_2,
    cmf3_3,
    hypergeometric_derived_2F1,
    hypergeometric_derived_3F2,
)

__all__ = [
    "CMF",
    "DFinite",
    "FFbar",
    "pFq",
    "MeijerG",
    "e",
    "pi",
    "symmetric_pi",
    "zeta3",
    "var_root_cmf",
    "cmf1",
    "cmf2",
    "cmf3_1",
    "cmf3_2",
    "cmf3_3",
    "hypergeometric_derived_2F1",
    "hypergeometric_derived_3F2",
]



from dataclasses import dataclass
import math
import sympy as sp
from sympy.abc import x

from ramanujantools import GenericPolynomial


r"""
We say that a polynomial continued fraction (PCF) with functions $a(x), b(x)$ is in Euler's form,
if there are polynomial $h_1(x), h_2(x), f(x)$ such that

        $$b(x) = -h_1(x) h_1(x)$$
        $$f(x)a(x) = f(x-1)h_1(x) + f(x+1)h_2(x+1)$$

In case $f(x)=1$, we say that it has the trivial Euler form.

If a PCF has such a form, then its limit is
$ K_1^\infty \frac{b(n)}{a(n)} = h_2(0)*f_1(0)/f_1(0)  ( S^{-1} - 1 )^{-1} $
where
$ S = \sum_{k=0}^\infty \frac{ f(0)f(1) }{ f(k)f(k+1) } \prod_{i=1}^k \frac{ h_1(i) }{ h_2(i+1} } $
"""


def multi_subsets(elem_to_count: dict):
    """
    A generator returning all the decompositions of a multiset to two multisets.
    elem_to_count needs to be a dictionary from the distinct elements in the multisets to the number of times
    they appear there.

    For example, the multiset {{x, x, x, y, y, z}}, should be represented by the dictionary {x:3, y:2, z:1}
    """

    def _multi_subsets(inner_elem_to_count: dict, keys, current1, current2):
        if len(keys) == 0:
            yield current1, current2
            return

        key = keys[0]
        total = inner_elem_to_count[key]
        for amount in range(total + 1):
            yield from _multi_subsets(
                inner_elem_to_count,
                keys[1:],
                {**current1, key: amount},
                {**current2, key: total - amount},
            )

    yield from _multi_subsets(elem_to_count, list(elem_to_count.keys()), {}, {})


@dataclass
class EulerSolution:
    r"""
    A dataclass representing a polynomial continued fraction of the form :

        $b(x) = -h_1(x) h_1(x)$
        $f(x)a(x) = f(x-1)h_1(x) + f(x+1)h_2(x+1)$
    """

    h_1: sp.Poly
    h_2: sp.Poly
    a: sp.Poly
    b: sp.Poly
    f: sp.Poly

    def __post_init__(self):
        # make sure that all the parameters are actually polynomials
        self.h_1 = sp.Poly(self.h_1, x)
        self.h_2 = sp.Poly(self.h_2, x)
        self.a = sp.Poly(self.a, x)
        self.b = sp.Poly(self.b, x)
        self.f = sp.Poly(self.f, x)

    def __hash__(self):
        return hash(
            (
                tuple(self.h_1.all_coeffs()),
                tuple(self.h_2.all_coeffs()),
                tuple(self.a.all_coeffs()),
                tuple(self.b.all_coeffs()),
                tuple(self.f.all_coeffs()),
            )
        )

    @staticmethod
    def _poly_eq(poly1: sp.Poly, poly2: sp.Poly):
        """
        Use this equality in case the polynomials have non integer algebraic coefficients
        """
        coeffs1 = [sp.simplify(c) for c in poly1.all_coeffs()]
        coeffs2 = [sp.simplify(c) for c in poly2.all_coeffs()]
        if len(coeffs1) != len(coeffs2):
            return False
        return all(c1 == c2 for c1, c2 in zip(coeffs1, coeffs2))

    def __eq__(self, other):
        if not isinstance(other, EulerSolution):
            return False
        # f is determined up to scalar multiplication
        f_coef = self.f.all_coeffs()
        g_coef = other.f.all_coeffs()
        if len(f_coef) != len(g_coef):
            return False
        lead_f = f_coef[0]
        lead_g = g_coef[0]
        # check if f and g are the same up to scalar
        if any(
            lead_f * g_elem - lead_g * f_elem for f_elem, g_elem in zip(f_coef, g_coef)
        ):
            return False
        return (
            EulerSolution._poly_eq(self.h_1, other.h_1)
            and EulerSolution._poly_eq(self.h_2, other.h_2)
            and EulerSolution._poly_eq(self.a, other.a)
            and EulerSolution._poly_eq(self.b, other.b)
        )


class Coefficients:
    r"""
    Represents the coefficients of a polynomial, where at index i there is the coefficient of $x^i$.
    For all other indices, the returned coefficient value will be zero.
    """

    def __init__(self, p: sp.Poly):
        self.coef = list(reversed(p.all_coeffs()))

    def __repr__(self):
        return repr(self.coef)

    def __getitem__(self, index):
        if 0 <= index < len(self.coef):
            return self.coef[index]
        return 0


class EulerSolver:
    r"""
    A solver for the problem of given two polynomials $a(x), b(x)$, find polynomials $f(x), h_1(x), h_2(x)$
    such that:

        $b(x) = -h_1(x) h_1(x)$
        $f(x)a(x) = f(x-1)h_1(x) + f(x+1)h_2(x+1)$

    All the functions are static, in order to prevent using global functions.
    """

    @staticmethod
    def solve_for(a: sp.Poly, b: sp.Poly) -> list[EulerSolution]:
        r"""
        Given two polynomials, $a(x),b(x)$, searches for all solutions to
                $b(x)=-h_1(x)h_2(x)$,
                $f(x)a(x) = f(x-1)h_1(x) + f(x+1)h_2(x+1)$
        works under the assumption that $a(x),b(x)$ are integeral, and so are $h_1(x),h_2(x)$ in the solutions.
        Returns a list of all possible solutions.
        """
        # Make sure that the parameters are actually polynomials
        a = sp.Poly(a, x)
        b = sp.Poly(b, x)

        roots = b.all_roots()
        if len(roots) < b.degree():
            raise ValueError(f"Could not find all the roots for {b.expr}")

        # count multiplicity of roots
        roots_dict = {}
        for root in roots:
            if root in roots_dict:
                roots_dict[root] += 1
            else:
                roots_dict[root] = 1

        leading_coefficient = b.LC()

        return EulerSolver.solve_for_monic_decomposition(
            a, roots_dict, leading_coefficient
        )

    @staticmethod
    def solve_for_monic_decomposition(
        a: sp.Poly, b_roots: dict[int, int], leading_coefficient_b: int
    ) -> list[EulerSolution]:
        r"""
        Given $b(x)=leading_coefficient * \prod (x-root)$, finds all the solutions to
                $b(x)=-h_1(x)*h_2(x)$
                $f(x)a(x) = f(x-1)h_1(x) + f(x+1)h_2(x+1)$
        and returns a list of these solutions
        """

        solutions = []
        for s1, s2 in multi_subsets(b_roots):
            # Each s_i is a dict {root: multiplicity}.
            # s1 and s2 are the decomposition of the roots of b(x) to two polynomials.
            # In particular, len(s1), len(s2) is going to be their degrees.

            deg_a = sp.degree(a)
            deg_h1 = sum(s1.values())
            deg_h2 = sum(s2.values())
            d = max(deg_a, deg_h1, deg_h2)
            if (deg_a == d) + (deg_h1 == d) + (deg_h2 == d) < 2:
                # At least two of the degrees should be maximal, otherwise there is no solution
                continue

            monic_poly_1 = sp.Poly(
                math.prod((x - root) ** power for root, power in s1.items()), x
            )
            monic_poly_2 = sp.Poly(
                math.prod((x - root) ** power for root, power in s2.items()), x
            )

            # we now have
            #       h_1 = c_1*monic_poly_1,
            #       h_2 = c_2*monic_poly_2
            # where
            #       c_1*c_2 = -leading_coefficient
            # If there is a solution, and d = max{deg(a), deg(h_1), deg(h_2)}, then
            #               coef_d(a)=coef_d(h_1)+coef_d(h_2)

            if deg_a < d:  # so that deg_a < d = deg_h1 == deg_h2
                # then c1 + c2 = 0, and c1*c2 = -c1*c1 = -leading_coefficient
                if leading_coefficient_b < 0:
                    continue  # TODO: add support to non integer solutions
                c = sp.sqrt(leading_coefficient_b)

                solutions += EulerSolver.solve_for_decomposition(
                    a, c * monic_poly_1, -c * monic_poly_2
                )
                solutions += EulerSolver.solve_for_decomposition(
                    a, -c * monic_poly_1, c * monic_poly_2
                )
                continue

            # now d=deg_a >= deg_h1, deg_h2
            leading_coefficient_a = a.LC()

            if deg_h1 < deg_h2:  # and automatically deg_a = deg_h2 = d
                c2 = leading_coefficient_a
                c1 = -leading_coefficient_b / c2
                solutions += EulerSolver.solve_for_decomposition(
                    a, c1 * monic_poly_1, c2 * monic_poly_2
                )
                continue

            if deg_h2 < deg_h1:  # and automatically deg_a = deg_h1 = d
                c1 = leading_coefficient_a
                c2 = -leading_coefficient_b / c1
                solutions += EulerSolver.solve_for_decomposition(
                    a, c1 * monic_poly_1, c2 * monic_poly_2
                )
                continue

            # Left with the case of deg_a = deg_h_1 = deg_h_2 = d, so that
            # c1 + c2 = leading_coefficient_a
            # c1 * c2 = - leading_coefficient_b
            # so that they are solutions to
            # (c - c1)*(c - c2) = c^2 - leading_coefficient_a * c - leading_coefficient_b
            disc = sp.sqrt(leading_coefficient_a**2 + 4 * leading_coefficient_b)
            c1 = (leading_coefficient_a + disc) / 2
            c2 = (leading_coefficient_a - disc) / 2

            solutions += EulerSolver.solve_for_decomposition(
                a, c1 * monic_poly_1, c2 * monic_poly_2
            )
            if c1 != c2:
                solutions += EulerSolver.solve_for_decomposition(
                    a, c2 * monic_poly_1, c1 * monic_poly_2
                )

        return solutions

    @staticmethod
    def solve_for_decomposition(
        a: sp.Poly, h_1: sp.Poly, h_2: sp.Poly
    ) -> list[EulerSolution]:
        """
        Tries to find a polynomial f solving the equation
                f(x)a(x) = f(x-1)h_1(x) + f(x+1)h_2(x+1)
        Returns a list of such solutions, which can be empty if non exist.
        """
        degrees = EulerSolver.find_possible_degrees(a, h_1, h_2)
        if len(degrees) == 0:
            return []

        solutions = []
        for d in degrees:
            solution = EulerSolver.solve_for_decomposition_with_degree(a, h_1, h_2, d)
            if solution:
                solutions.append(solution)

        return solutions

    @staticmethod
    def find_possible_degrees(a: sp.Poly, h_1: sp.Poly, h_2: sp.Poly) -> list[int]:
        """
        Looking for the at most two possible degrees of a polynomial f which solves
                f(x)a(x) = f(x-1)h_1(x) + f(x+1)h_2(x+1)
        return these degrees in a list (which has size 0, 1, or 2).
        """

        d = max(sp.degree(a), sp.degree(h_1), sp.degree(h_2))

        # Create lists of the coefficients of size d+1, such that p[i] will be the i-th coefficient of the polynomial p.
        a_coef = Coefficients(a)
        h_1_coef = Coefficients(h_1)
        h_2_plus_coef = Coefficients(sp.Poly(h_2.subs(x, x + 1), x))

        # Condition 1: Leading coefficients are equal.
        if a_coef[d] != h_1_coef[d] + h_2_plus_coef[d]:
            return []

        # Condition 2: The second leading coefficients are equal.
        #              The degree d_f appears in this equation, so unless it is 0=0, we can find d_f.
        num = a_coef[d - 1] - h_1_coef[d - 1] - h_2_plus_coef[d - 1]
        denom = h_2_plus_coef[d] - h_1_coef[d]
        if denom != 0:
            deg = float(sp.simplify(num / denom))
            if deg.is_integer() and int(deg) >= 0:
                return [int(deg)]
            return []

        if num != 0:
            return []

        # Condition 3: The third leading coefficients are equal. In case the previous condition didn't provide
        #              the degree, then we get here a nontrivial quadratic equation.
        aa = -(h_1_coef[d] + h_2_plus_coef[d])
        bb = h_1_coef[d - 1] - h_2_plus_coef[d - 1]
        cc = a_coef[d - 2] - h_1_coef[d - 2] - h_2_plus_coef[d - 2]

        coef2 = aa * 0.5
        coef1 = bb - aa * 0.5
        coef0 = cc

        # the possible degrees are the roots for coef2 * x^2 + coef1 * x + coef0 = 0
        disc_squared = coef1**2 - 4 * coef2 * coef0
        if disc_squared < 0:
            return []
        disc = math.sqrt(coef1**2 - 4 * coef2 * coef0)
        roots = [
            sp.simplify((-coef1 + disc) / (2 * coef2)),
            sp.simplify((-coef1 - disc) / (2 * coef2)),
        ]

        return [int(root) for root in roots if float(root).is_integer() and root >= 0]

    @staticmethod
    def solve_for_decomposition_with_degree(
        a: sp.Poly, h_1: sp.Poly, h_2: sp.Poly, d_f: int
    ) -> EulerSolution | None:
        """
        Tries to find a polynomial f of degree d_f solving the equation
                f(x)a(x) = f(x-1)h_1(x) + f(x+1)h_2(x+1)
        If there is such a polynomial returns the solution, otherwise returns None.
        """

        b = -h_1 * h_2
        h_2_plus = sp.Poly(h_2.subs(x, x + 1), x)

        if d_f == 0:
            if a - (h_1 + h_2_plus) == 0:  # The only solution for constant polynomials
                return EulerSolution(
                    h_1=sp.Poly(h_1, x), h_2=sp.Poly(h_2, x), a=a, b=b, f=sp.Poly(1, x)
                )
            return None

        # Create the polynomial f, and the recursion it needs to solve, namely
        #    f(x)a(x) = f(x-1)h_1(x) + f(x+1)h_2(x+1)
        f, f_ = GenericPolynomial.of_degree(deg=d_f, var_name="f_", s=x, monic=True)
        f_plus = sp.Poly(f.subs(x, x + 1), x)
        f_minus = sp.Poly(f.subs(x, x - 1), x)

        p = f * a - f_minus * h_1 - f_plus * h_2_plus

        system = [sp.simplify(coefficient) for coefficient in p.all_coeffs()]
        system = [equation for equation in system if equation != 0]
        if len(system) == 0:  # no conditions on f_
            return EulerSolution(
                h_1=sp.Poly(h_1, x), h_2=sp.Poly(h_2, x), a=a, b=b, f=sp.Poly(f, x)
            )

        assignment = sp.solve(p.all_coeffs(), f_)
        if len(assignment) == 0:
            return None

        # TODO: If there is a solution, does it have to be unique?
        f_solved = f.subs(assignment)
        return EulerSolution(
            h_1=sp.Poly(h_1, x), h_2=sp.Poly(h_2, x), a=a, b=b, f=sp.Poly(f_solved, x)
        )
